; Movimento Led Utilizzando le Tabelle e un Ritardo.
;
; +------------------------------+
; | Author: Matteo Bertozzi      |
; | Release Date: 30 Jan 2005    |
; +------------------------------+
; Responsabilita: Solo le Istruzioni in MAIUSCOLO
;                 sono state scritte da Matteo Bertozzi.

; -----------------------------------------------------------------------
; RACCONTO DELL'ESERCIZIO
; Devo far muovere i led Rossi (PDA) in un modo abbastanza figo.
; Pero' mi tocca farlo a mano (Se qualcuno mi trova un modo utilizzando
; cicli e operazioni varie gli pago la cena). Per questo motivo
; Utilizziamo le Tabelle che non sono altro che delle operazioni 
; fatte quasi a mano per settare i led secondo lo schema che vogliamo.
;
; Le Tabelle utilizzate seguiranno questo schema:
; tableX:
;		...		; SET dei Led
;	SJMP tableX		; Loop Infinito sulla TabellaX
;
; Il Settaggio dei led viene fatto in questo modo
;	MOV A,#11111111b	; Setto il valore dei LED
;	MOVX @DPTR,A		; Lo Invio ai led DPTR contiene l'Ind. PDA
;	ACALL ritardo		; Eseguo un ritardo
;
; Il ritardo mi serve per vedere gli effetti dei led che si accendono
; in sequenza o stupidate simili.
; -----------------------------------------------------------------------

.org 8050h		; Posizione di Start del PGM in MEM

; +-----------+----------+---------------------------------------------+
; | Indirizzo | Registro | Descrizione                                 |
; +-----------+----------+---------------------------------------------+
; |   0FA00h  |    PDA	 | Registro dati del port A (Led Rossi)        |
; |   0FA01h  |    PDB	 | Registro dati del port B                    |
; |   0FA02h  |    PDC	 | Registro dati del port C (Led Gialli/Verdi) |
; |   0FA03h  |    CNT	 | Registro di controllo                       |
; +-----------+----------+---------------------------------------------+
;
; :: Informazioni sui LED ::
;    Bit = 0 -> LED ON
;    Bit = 1 -> LED OFF
;    8 Led, 8bit -> 01111111b = Solo il Primo Acceso
; 

; C O S T A N T I
; Spero che il .equ sia come un #define che sostituisca
; in tempo di assemblazione cosi' ho la possibilita' di 
; usarlo con #CONST e @CONST.
.equ INDPDA, 0FA00h	; Indirizzo Registro dati del port A, Led Rossi
.equ INDCNT, 0FA03h	; Indirizzo Registro di controllo
.equ CYCLSLEEP,240	; Numero di Cicli del Ritardo 
; END COSTANTI

SJMP main		; Salto alla Funzione Principale

; Funzione di Ritardo per i led
; 2 Cicli Annidati con una NOP. Settare la Costante CYCLSLEEP
; per definire il ritardo (l'ho fatto a spanne, non e' che sia
; una gran cosa)
; DJNZ: DECREMENT Register and JMP IF NOT ZERO
ritardo:
	MOV R0,#CYCLSLEEP
	rlp1:
		MOV R1,#CYCLSLEEP
		rlp2:
			NOP
		DJNZ R1,rlp2
	DJNZ R0,rlp1
	RET

; Primo schema di Movimento LED - Led Che si Incontrano
; ATTENZIONE: non e' una funzione... e' solo un salto
;             nel codice per avere piu' "disegni".
;             quindi non si usa la 'RET'.
table1:
	MOV A,#11111111b	; Led Spenti
	MOVX @DPTR,A
	ACALL ritardo
	
	MOV A,#01111110b	; Accesi i due Led Esterni
	MOVX @DPTR,A
	ACALL ritardo
	
	MOV A,#10111101b
	MOVX @DPTR,A
	ACALL ritardo
	
	MOV A,#11011011b
	MOVX @DPTR,A
	ACALL ritardo
	
	MOV A,#11100111b	; Accesi i due Led Centrali
	MOVX @DPTR,A
	ACALL ritardo
	
	MOV A,#11011011b	; Inizio Ritorno
	MOVX @DPTR,A
	ACALL ritardo
	
	MOV A,#10111101b
	MOVX @DPTR,A
	ACALL ritardo
	
	MOV A,#01111110b	; Accesi i due Led Esterni
	MOVX @DPTR,A
	ACALL ritardo
	
	SJMP table1		; Looppa in Infinito sulla Table1

; SUPERCAR - Forse
; t2lpup: Accendo i led in sequenza da 11111111b a 00000000b
; t2lpdw: Spengo i led in sequenza da 00000001b a 11111111b
; 0: Acceso - 1: Spento
table2:
	MOV R2,#8			; Counter 8 bit dei led	
	MOV A,#11111111b
	t2lpup:			
		MOVX @DPTR,A
		ACALL ritardo
		MOV B,#00000010b	; Per la Divisione per 2
		DIV AB			; Divido A per 2 (00000010b)
		DJNZ R2,t2lpup
	
	ACALL ritardo			; Aspetto un attimo prima di Tornare
	
	MOV R2,#8			; Counter 8 bit dei led	
	MOV A,#00000001b		; Spento solo L'ultimo
	t2lpdw:
		MOVX @DPTR,A
		ACALL ritardo
		MOV B,#00000010b	; Per la Moltiplicazione per 2
		MUL AB
		DJNZ R2,t2lpdw
	SJMP table1			; Looppa in Infinito sulla Table1

; Main Progs
; - Setto lo Stack per le chiamate a funzione
; - Setto il CNT
; - Setto il PDA (Led Rossi)
; - Salto al Main Loop (table1)
main:
	mov	sp,#8Fh		; Imposta lo stack per 'ritardo()'
	
	; SET CNT
	MOV	A,#10001010b	; Imposto la Porta di Controllo
	MOV	DPTR,#INDCNT	; Nel DPTR inserisco l'indirizzo del CNT
	MOVX	@DPTR,A		; Nel @DPTR metto A
	
	; SET PDA (Led Rossi)
	MOV	DPTR,#INDPDA	; Nel DPTR inserisco l'indirizzo del PDA

	SJMP	table1		; Eseguo la Configurazione del LED 1
	;SJMP	table2		; Eseguo la Configurazione del LED 2

.end	; Servira' a qualcosa ?

; -------------------------------
; Tabs width: "8"
; Editor Font: "Monospace 12"
; -------------------------------
