/***************************************************************************
 *  i8052.h
 *
 *  Sun Nov 14 10:49:20 2004
 *  Copyright  2004 Ing. Stefano Salvi
 *  Email stefano@salvi.mn.it
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef _I8052_H
#define _I8052_H

#ifndef _SFRMEMORY_H
#include "sfrmemory.h"
#endif
#ifndef OPCODE_H
#include "opcode.h"
#endif
#ifndef _ERRORINTERFACE_H
#include "errorinterface.h"
#endif

class DataMemory;
class XDataMemory;
class CodeMemory;
class Serial_Container;

class hexListFile {
	FILE *lst;
	FILE *hex;
	char *src;
	unsigned curad;
	unsigned lstPending;
	unsigned hexPending;
	unsigned char buf [16];
	
	char *changeExt (const char *name, const char *ext);
	void flush ();
public:
	hexListFile () { lst=NULL; hex=NULL; src = NULL; curad = lstPending = hexPending = 0; };
	~hexListFile ();
	bool open (const char *name);
	void add (unsigned ad, unsigned char val);
	void newLine (char *l);
	void error (char *e);
};

class i8052 {
friend class opcContent;
	static parseNode *parseTree;
	static symtabNode *genSymTab;
	Serial_Container *interface;
	volatile enum stat_val status;
	// Aree di memoria
	DataMemory *dmemory;
	XDataMemory *xmemory;
	CodeMemory *cmemory;
	SFRMemory *SFRmemory;

	pthread_t Emulate_thread;
	static void *Emulate_wrapper (void *th);
	void *emulate (void);
	
	int realStep ();
	void setStatus (stat_val s);
	bool pass (const char *source, bool second, errorInterface *msg, hexListFile *f);
	void printError (bool warning, int linnum, errorInterface *msg, hexListFile *f);
	bool cmpSym (char *&s,const char *d);

	static int bitAddress (char *&s, int num, bool &bitAddr, bool second);
	static int val (char *&s, bool &bitAddr, unsigned pc, bool second);
	static int unexpr (char *&s, bool &bitAddr, unsigned pc, bool second);
	static int mulexpr (char *&s, bool &bitAddr, unsigned pc, bool second);
	static int addexpr (char *&s, bool &bitAddr, unsigned pc, bool second);
	static int GetInt (char *&a);
	static int expr (char *&s, bool &bitAddr, unsigned pc, bool second);
	static char *skipBlanks (char *s);
	static char *skipSymbol (char *s);
	static bool addSymbol ( char *str, int v) { return symtabNode::addToRoot (genSymTab, str, v); };
	static void clearSymTab () {if (genSymTab) { delete genSymTab; genSymTab = NULL;}};

protected:
	static OpCode IstSet [256];
public:
	DataMemory *GetData () { return dmemory; }
	XDataMemory * GetXData () { return xmemory; };
	CodeMemory * GetCode () { return cmemory; };
	SFRMemory * GetSFR () { return SFRmemory; };
	
	unsigned GetMnemonic (int pc, char *mnemonic);
	int Asm (char *buf,unsigned Pc);
	int commonAsm (char *buf,unsigned Pc, unsigned char *code, opcContent *&ists, bool secondpass);
	void assembler (const char *source, errorInterface *msg);

	void receiver (char c) { SFRmemory->receiver (c); };
	void doUpload (char *file);
	void stop () { setStatus (stopped); };
	void run (int where);
	void runPc () { run (SFRmemory->GetPc());};
	bool isEmulating () { return status == emulating;};
	bool isRunning () { return status != stopped; };
	void setRunning (bool v) { status = (v) ? running : stopped; };
	void redrawIfChanged ();
	void reload ();
	
	void setPPI (ppiinterface *c_PPI);

	void step ();
	
	i8052 (Serial_Container *i);
};

#endif /* _I8052_H */
