/***************************************************************************
 *  memory.h
 *
 *  Sun Nov 14 10:49:20 2004
 *  Copyright  2004 Ing. Stefano Salvi
 *  Email stefano@salvi.mn.it
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
/*
 * Contenitori di Memoria:
 * Memory -> generica area di memoria
 * |-> DataMemory -> Memoria DATA
 * |-> XDataMemory -> Memoria XData
 *     |-> CodeMemory -> Memoria Code
 * Pannelli di visualizzazione
 * MemoryPanel -> Base dei pannelli di visualizzazione (serve per Data)
 * |-> XMemoryPanel -> Pannello di visualizzazione XData (si preoccupa di caricare i dati)
 * |-> CMemoryPanel -> Oltre al caricamento dei dati, si occupa dell'assembler
 * Finestre principali
 * Data -> Area DATA
 * XData -> Area XDATA
 * Code -> Area CODE
 *
 * Status:
 * Stopped: la scheda e l'emulatore sono spenti.
 *					- Le richieste di "reload" sono gestite
 *					- Le ricieste di "redraw" sono gestite
 * Running:	La scheda è in funzione (l'emulatore no)
 *					- Le richieste di "reload" sono bloccate
 *					- Le richieste ri "redraw" sono eseguite
 * Emulating: La scheda non è conessa, PCMON sta emulando
 *					- Le richieste di "reload" vengono eseguite
 *					- Le richieste di "redraw" vengono accodate
 */

#ifndef MEMORY_H
#define MEMORY_H

#ifndef _UPDATABLE_H
#include "updatable.h"
#endif

/* Area di memoria (da specializzare come "data", "Code" o "XData"
 */
enum stat_val { stopped, running, emulating};
 
class Memory {
friend class CodeMemory;
friend class HGPRegister;
protected:
	Updatable *r;
	unsigned char *data;
	unsigned start;
	unsigned size;
	enum stat_val status;
	bool toRedraw;
public:
	virtual void SetDato (unsigned ad,char val);
	virtual unsigned char GetDato (unsigned ad);
	void setRefresh (Updatable *newR) { r = newR; };
	unsigned getSize () { return size; };
	unsigned getStart () { return start; };
	unsigned getEnd () { return start+size; };
	
	virtual void setStatus (stat_val s) { status = s;};

	virtual void reload () = 0;				// Ricarica l'intero contenuto dello spazio
	virtual void saveValue (int pos) = 0;	// Registra il dato sulla scheda
	virtual void redraw () { if (status == emulating) { toRedraw = true; } else { r -> redraw (); } };
	virtual void redrawIfChanged () { if (toRedraw) { toRedraw = false; r -> redraw ();} };

	Memory (int c_start, int c_size);
	Memory (unsigned char *d, unsigned c_start, unsigned c_size) { data=d; start=c_start, size=c_size; };
  	unsigned char *getData (unsigned base) { return data + base; };
};

#endif
//
