/***************************************************************************
 *  opcode.h
 *
 *  Sun Nov 14 10:49:20 2004
 *  Copyright  2004 Ing. Stefano Salvi
 *  Email stefano@salvi.mn.it
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef OPCODE_H
#define OPCODE_H

#ifndef toupper
#include <ctype.h>
#endif

#ifndef NULL
#define	NULL	0
#endif

#ifndef SYMTAB_H
#include "symtab.h"
#endif
	
class DataMemory;
class XDataMemory;
class Memory;
	
// abbiamo 58 diversi opcode per la "mov"
// Se pero' decodifichiamo il nimero del registro, dove c'e' un
// registro, si devono togliere 40 codici, scendendo quindi a 18.
// Se inserisco stringhe di 58 opcode per ogni nodo,
// Arrivo ad un massimo di 14848 bytes.
// Se inserisco ogni opcode in una lista linkata,
// ottengo 45 * 4 + 256 * 5 = 1460
// 45 opcode, 192 caratteri, 3,26 caratteri in media (sicuramente meno di 21 per lista)
// 89 nodi massimo (sicuramente molti meno) , 5196 caratteri. Non e' una gran quantita'
// 14 -> primo livello
// 35 -> secondo livello
// 45 -> terzo livello
// 94 = totale
#define	MAXOPCODE	19

#define	issymbase(c) (((c)>='a' && (c) <= 'z') || ((c)>='A' && (c) <= 'Z') || (c) == '_')
#define	issym(c) (issymbase(c) || (c) == '.')
#define	issym_cont(c) (issymbase(c) || isdigit(c))
	
enum nodeType { opcode, org, equ, db, dw, df, skip , end};

class opcContent {
	int nopcode;
	unsigned char opcodeList [MAXOPCODE];
	nodeType type;
public:
	opcContent () { nopcode = 0; };
	void add (unsigned char code, nodeType t);
	int getNum () { return nopcode;};
	nodeType getType () { return type; };
	unsigned char operator [] (int i) { return opcodeList [i]; };
};

class symContent {
	bool	present;
	int	value;
public:
	symContent () { present = false; value = 0; };
	bool set (int v) { if (present) { return false; } else { value = v; present = true; } return true; };
	int get () { return value; };
	bool isPresent () { return present; };
};

class Node {
	char c;
	Node *nextChar;	// Prossimo carattere allo stesso livello
	Node *nextLevel;	// Prossimo carattere nel simblo;
protected:
	virtual Node *New (char nC, Node *next) = 0;
	Node *add (Node **parseTree, char *str);
public:
	Node (char nC, Node *next) { nextChar = next; nextLevel = NULL; c = toupper (nC); };
	virtual ~Node () {  if (nextChar) { delete nextChar; } if (nextLevel) { delete nextLevel; } };
	Node *find (char *str);
};

class parseNode : public Node {
	opcContent content;
	void setContent (nodeType type, unsigned char opc);
protected:
	virtual Node *New (char nC, Node *next) { return new parseNode (nC, next); };
public:
	static int maxopcode;
	static int nnodes;

	parseNode (char nC, Node *next) : Node (nC, next) { nnodes++; }
	static void addToRoot (parseNode *&parseTree, char *str, nodeType type, unsigned char opc);
	opcContent *find (char *str);
};

class symtabNode : public Node {
	symContent content;
protected:
	virtual Node *New (char nC, Node *next) { return new symtabNode (nC, next); };
public:
	symtabNode (char nC, Node *next) : Node (nC, next) { }
	static bool addToRoot (symtabNode *&symTab, char *str, int v);
	symContent *find (char *str);
};

/*
- Primo argomento
1,$s/\( @R.*",OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*,OpCode::.*\)/\1indReg\2/
1,$s/\( R.*",OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*,OpCode::.*\)/\1reg\2/
1,$s/\( %s.*",OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*,OpCode::.*\)/\1dir\2/
1,$s/\("[A-Z]* A.*",OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*,OpCode::.*\)/\1aReg\2/
1,$s/\("[A-Z]* AB",OpCode::.*,OpCode::.*,OpCode::\)aReg\(,OpCode::.*,OpCode::.*\)/\1bReg\2/
1,$s/\("[A-Z]* C.*",OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*,OpCode::.*\)/\1cReg\2/
1,$s/\("[A-Z]* DPTR.*",OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*,OpCode::.*\)/\1dptrReg\2/
1,$s/\("[A-Z]* @DPTR.*",OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*,OpCode::.*\)/\1indDptr\2/
1,$s/\("[A-Z]* @A+DPTR.*",OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*,OpCode::.*\)/\1idxDptr\2/
1,$s/\("[A-Z]* @A+PC.*",OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*,OpCode::.*\)/\1idxPc\2/
- Secondo argomento
1,$s/\(,@A+PC.*",OpCode::.*,OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*\)/\1idxPc\2/
1,$s/\(,@A+DPTR.*",OpCode::.*,OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*\)/\1idxDptr\2/
1,$s/\(,@DPTR.*",OpCode::.*,OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*\)/\1indDptr\2/
1,$s/\(,A.*",OpCode::.*,OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*\)/\1aReg\2/
1,$s/\(,@R.*",OpCode::.*,OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*\)/\1indReg\2/
1,$s/\(,R.*",OpCode::.*,OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*\)/\1reg\2/
1,$s/\(,#%s.*",OpCode::.*,OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*\)/\1Pimm\2/
1,$s/\(,%s",OpCode::.*,OpCode::.*,OpCode::.*,OpCode::\)noParg\(,OpCode::.*\)/\1dir\2/
- Terzo argomento
1,$s/\(,[^,]*,%s",OpCode::.*,OpCode::.*,OpCode::.*,OpCode::.*\,OpCode::.*\)noParg/\1dir/
 */

#define nonfatal	noop
class OpCode {
public:
	enum ErrVals { noerr, toobig, toofar, isbitaddr, noopcode, noop, numerr, charconst, numerr2, nosfr, parserr, dupsym, nosym, strerr, idxerr, indreg, inderr,
		bitnumerr, notbitadressable, parnotbal, illbitaddrop, divbyzero, illoper };
	enum argType { noarg, imm, data, bit, imm16, reladdr, fixedarg, absaddr, paddr80 = 0x80, paddr81 = 0x81, paddr82 = 0x82,
		paddr83 = 0x83, paddr84 = 0x84, paddr85 = 0x85, paddr86 = 0x86, paddr87 = 0x87 };
	enum argParseType { noParg, Pimm, dir, bitDir, notBitDir, reg, indReg, indDptr, idxPc, idxDptr, dptrReg, aReg, cReg, bReg };
	struct argStore {
		argParseType type;
		int val;
	};
	
	static SymTab Syms [];
private:
	static ErrVals ErrCode;
	static char *ErrStrs [];

	char *Mnemonic;
	argType Args [2];
	argParseType pars [3];
	unsigned char Len;
	unsigned char Cycles;

	static DataMemory *Data;
	static XDataMemory *Xdata;
	static Memory *Code;

	void MkArg (char *a,int offs,unsigned Ad,argType Arg);
	unsigned GetArgVal (int offs,unsigned Ad,argType Arg);
	unsigned char Oper (unsigned char Opcode,
		unsigned char Op1, unsigned char Op2);

public :
	OpCode (char *o, argType A1, argType A2, argParseType P1, argParseType P2, argParseType P3, int L, int C) 
		{ Mnemonic = o; Args [0] = A1; Args [1] = A2; pars [0] = P1; pars [1] = P2; pars [2] = P3; Len = L; Cycles = C; };
	static void setProcessor (DataMemory *data, XDataMemory *xdata, Memory *code) 
		{ Data = data; Xdata = xdata; Code = code; };
	int GetLen (void) { return Len; };
	int getCycles () { return Cycles; };
	unsigned GetMnemonic (unsigned Ad,char *s);
	unsigned NextIst (unsigned char *p,unsigned Pc,int Calls);
	static char * const ErrStr (ErrVals code);
	static void ResetErr (void) { ErrCode = noerr; };
	static ErrVals GetErr (void) { return ErrCode; };
	static void SetErr (ErrVals newErr) { ErrCode = newErr;};
	int Asm (int num,argStore *pars,unsigned Ad, unsigned char *a, bool secondpass = true);
	unsigned Exec (unsigned Pc);
	void add (parseNode *&parseTree, unsigned char opc);
};

#endif
