/***************************************************************************
 *  pcmon.h
 *
 *  Sun Nov 14 10:49:20 2004
 *  Copyright  2004 Ing. Stefano Salvi
 *  Email stefano@salvi.mn.it
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef PCMON_H
#define PCMON_H

#include "termdrawing.h"
#include "dialogs.h"
// #include <pthread.h>
// #include <termios.h>

#ifndef _SERIAL_H
#include "serial.h"
#endif

class Pc;
class Sfr;
class PcData;
class XData;
class Code;
class EditWindow;
class monWindow;
class ppiwin;
class errorInterface;

#ifndef _I8052_H
class i8052;
#endif
	
class Serial;

class Pcmon : public Gtk::Window, public Serial_Container {
  friend class EditWindow;
  friend class Pc;
  friend class Sfr;
  friend class Register;
  friend class Code;
  friend class MemoryPanel;
  friend class CMemoryPanel;
  friend class Serial;

  SigC::Connection EmulTimerId;	/* "timer" signal (txt widget) */

  Gtk::VBox vbox;
  EditWindow *edits;
  Gtk::HBox status;
  Gtk::Frame ConnFrame;
  Gtk::Frame RunFrame;
  Gtk::Label Message;

  Glib::RefPtr<Gtk::Action::Action> stopButton;
  Glib::RefPtr<Gtk::UIManager> m_refUIManager;
  Glib::RefPtr<Gtk::ActionGroup> m_refActionGroup;
  Gtk::Menu* menu_finestre;
  
  /* ui_merge_id */ int windowUiMergeId;
  Glib::RefPtr<Gtk::ActionGroup> win_refActionGroup;

  SigC::Connection FileButtonConnection;

  int quit ();
  gint selectionclosed (GdkEventAny*);

  void menuAndToolbarBuild (void);

  virtual bool on_key_press_event (GdkEventKey* p1);
  virtual bool on_delete_event (GdkEventAny *e);
  void setupCom ();

  // Funzioni dei menu (private)
  void menuUpload();
  void menuShow (monWindow *toShow);

  bool tick();
protected:

  terminal_drawing_area drawing_area;
  Gtk::Label Connected;
  Gtk::Label Running;

  i8052 *processor;

  // Finestre del monitor
  Sfr *sfr;
  Pc *pc;
  PcData *data;
  XData *xdata;
  Code *code;
  ppiwin *PPIWin;

  void ceateStockItem (const char* const* data, const Glib::ustring& id, const Glib::ustring& label);
  void doRun (int where);

public:
	// Interfaccia Serial_Container
	virtual void reload ();
	virtual bool isRunning ();
	virtual void setRunning (bool v);
	virtual void setConnectedMessage (const char*s);
	virtual void setRunningMessage (const char*s);
	virtual void setMessage (const char*s);
	virtual void tPutc (char c);
	virtual void startEmulation ();
	// Fine Interfaccia Serial_Container

	void windowMenu ();
	
	// Funzioni dei menu (pubbiliche)
	void menuQuit ();
	void menuOpen ();
	void menuNew ();
	void menuRun ();
	void menuRunPc ();
	void menuRunStart ();
	void menuStep ();
	void menuStop ();
	void doUpload (char *file);
	
	void delEditor (EditWindow *e);
	void assembler (const char *source, errorInterface *msg);
	
	Pcmon(char *s);
};

#endif
