/***************************************************************************
 *  cmemorypanel.cc
 *
 *  Sun Nov 14 10:49:20 2004
 *  Copyright  2004 Ing. Stefano Salvi
 *  Email stefano@salvi.mn.it
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include <glib/gmacros.h>
#include <glib.h>
#include <gtkmm.h>
#include <libgnomemm.h>

#include <iostream>	// Per la stampa dell'eventuale errore
 
#include "pcmon.h"
#include "cmemorypanel.h"
#include "memory.h"
#include "pc.h"

void CMemoryPanel::show (unsigned address)
{
	position (address, false);
}

void CMemoryPanel::position (unsigned address, bool exact)
{
	if (address >= data->getStart() && address < data->getEnd()) {
		curY = address - data -> getStart ();
		// Visualizza la posizione del cursore
		if (curY < yOffs) {
			VAdj->set_value (curY);
		} else if (curY >= yOffs + nLines - 2) {
			if (exact || curY > yOffs + 2 * nLines - 2) {
				VAdj->set_value (curY);
			} else {
			unsigned nextOffs = (Ists[curY - nLines/2 - yOffs + 1][0]) ? curY - nLines/2 + 1 : 
					Next [curY - nLines/2 - yOffs + 1] - data -> getStart ();
	    		VAdj->set_value (nextOffs);
			}
		}
		redraw ();
	}
}

void CMemoryPanel::menuRunTo ()
{
	sfr -> SetPc (curY + data -> getStart ());
	processor -> runPc ();
}


void CMemoryPanel::showLine (Glib::RefPtr<Gdk::Window> win, unsigned line, int y)
{
Glib::RefPtr<Gdk::GC> gc;
unsigned i;
char buf [80];
unsigned address = line + data -> getStart ();

	if (address == sfr -> GetPc ()) {
		gc = get_style()->get_dark_gc(Gtk::STATE_ACTIVE);
	} else if (sfr -> isBp (address)) {
		gc = get_style()->get_mid_gc(Gtk::STATE_ACTIVE);
	} else {
		gc = get_style()->get_white_gc();
	}
	win->draw_rectangle (gc,true, 0, y,get_width(), text_h);

	gc = get_style()->get_black_gc();
	Glib::RefPtr<Pango::Layout> pangolayout = create_pango_layout("");
	pangolayout->set_font_description  (*font); 
	if (line < data -> getSize ())
	{
	unsigned char dato = data->GetDato(address);
		sprintf (buf ," %04X %02X ",address, dato & 0xff);
		if (dato >= ' ' && dato < 127) {
			sprintf (buf + CHARPOS ,"'%c' ",dato);
		} else {
			sprintf (buf + CHARPOS ,"    ");
		}
		if (line - yOffs < nLines) {
			strcpy (buf + ASMPOS, (line == curY && IstChanged) ? EdBuf : Ists [line - yOffs]);
		}

		i = strlen (buf);
		if (xOffs < i) {
			i = xOffs;
		}

		pangolayout -> set_text (buf + i);
		win->draw_layout(gc, border , y, pangolayout);

		// Cursore
		if (curOn && curY == line) {
			switch (Zone)
			{
				case 0:	// Hex
					i = (low) ? 7 : 6;
					break;
				case 1:	// Ascii
					i = 10;
					break;
				case 2:	// Assembler
					i = 13 + curX;
					break;
			}
			win->draw_rectangle (gc,true,(curPos () - xOffs) * text_w + border, y + text_h - 2, text_w, 2);
		}
	}
}

void CMemoryPanel::ChangeIst (void)
{
	if (!IstChanged) {
		strcpy (EdBuf,Ists [curY - yOffs]);
		CIstLen = strlen (EdBuf);
		IstChanged = true;
	}
}

void CMemoryPanel::key (int k)
{
GdkRectangle redraw_rect;
unsigned nextOffs;
	curOn = true;
	tick();	// Spegne il cursore
	switch (k)
	{
		case GDK_Up :  // Freccia Su
			if (Zone == 2)
			{
				if (Asm ()) break;      // l'ASM e' sbagliato
				curY = Prev [curY - yOffs] - data -> getStart ();
			}
			else if (curY) curY --;
      			curX = 0;
      		break;
		case GDK_Down:  // Freccia Giu
			if (Zone == 2) {
				if (Asm ()) break;      // l'ASM e' sbagliato
				curY = Next [curY - yOffs] - data -> getStart ();
			} else  {
				curY ++;
			}
			if (curY >=  data -> getSize ()) {
				curY = data -> getSize ()  - 1;
			}
			curX = 0;
			break;
		case GDK_Right :  // Freccia Destra
			switch (Zone)
			{
				case 0 :        // Hex
					if (!low) low = true;
					else
					{
						low = false;
						if (curY < data -> getSize ()) {
							curY ++;
						}
						curX = 0;
					}
					break;
				case 1 :        // Char
					if (curY < data -> getSize ()) {
						curY ++;
					}
					curX = 0;
					break;
				case 2 :        // Asm
 					ChangeIst ();
					if (curX < CIstLen) {
						curX ++;
					}
					break;
			}
			break;
		case GDK_Left :  // Freccia Sinistra
			switch (Zone)
			{
				case 0 :        // Hex
					if (low) {
						low = false;
					}
					else
					{
						low = true;
						if (curY) {
							curY --;
						}
						low = 0;
					}
					break;
				case 1 :        // Char
					if (curY) {
						curY --;
					}
					curX = 0;
					break;
				case 2 :        // Asm
					if (curX) {
						curX --;
					}
					break;
			}
			break;
		case GDK_Page_Up :  // Pagina Su
			if (Zone == 2 && Asm ()) {	// l'ASM e' sbagliato
				break;
			}
			if (curY >  nLines / 2) {
				curY -= nLines / 2;
			} else  {
				curY = 0;
			}
			if (yOffs > nLines / 2) {
				yOffs -= nLines / 2;
			} else {
				yOffs = 0;
			}
			VAdj->set_value (yOffs);
			curX = 0;
			break;
		case GDK_Page_Down :  // Pagina Giu
			if (Zone == 2 && Asm ()) break; // l'ASM e' sbagliato
			curY += nLines / 2;
			if (curY >= data -> getSize ()) {
				curY = data -> getSize () - 1;
			}
			nextOffs = yOffs +nLines / 2;
			if (nextOffs >= data -> getSize () - nLines)  {
				nextOffs = data -> getSize () - nLines;
			}
			if (Ists[nextOffs - yOffs][0] == 0) {
				nextOffs = Next [nextOffs - yOffs] - data -> getStart ();
			}
			VAdj->set_value (nextOffs);
			curX = 0;
			break;
		case GDK_Home :  // Home
			if (Zone == 2 && Asm ()) break; // l'ASM e' sbagliato
			low = false;
			curX = 0;
			curY = 0;
			break;
		case GDK_End :  // End
			if (Zone == 2 && Asm ()) break; // l'ASM e' sbagliato
			low = true;
			curX = 0;
			curY = data -> getSize () - 1;
			break;
		case GDK_Tab :
			if (Zone == 2 && Asm ()) break; // l'ASM e' sbagliato
			Zone++;
			if (Zone == 2 && Ists [curY-yOffs][0]==0) {
				Zone = 0;
			}
			if (Zone > 2) {
				Zone = 0;
			}
			break;
		case GDK_ISO_Left_Tab:	// Backtab
			if (Zone) Zone --;
			else Zone = 2;
			if (Zone == 2 && Ists [curY-yOffs][0]==0) {
				Zone = 1;
			}
			break;
		case 0x1b :	// Escape
		case GDK_Escape :	// Escape
			IstChanged = 0;
			RefreshLines (1);
			curX = 0;
			break;
		case GDK_Delete :	//  0xFFFF
			if (Zone == 2)
			{
 				ChangeIst ();
				if (curX < CIstLen)
				{
					strcpy (EdBuf + curX,EdBuf + curX + 1);
					CIstLen --;
				}
				RefreshLines (1);
			}
			break;
		case GDK_BackSpace :	// Back
			if (curX && Zone == 2)
			{
				curX --;
				strcpy (EdBuf + curX,EdBuf + curX + 1);
				CIstLen --;
				RefreshLines (1);
			}
			break;
		default :
			redraw_rect.x = (5 - xOffs) * text_w + border;
			redraw_rect.y = (curY - yOffs) * text_h + border;
			redraw_rect.width = 44 * text_w;
			redraw_rect.height = text_h;

			switch (Zone)
			{
				case 0 :        // Hex
					if ((k >= '0' && k <= '9') || (k >= 'A' && k <= 'F') ||
						(k >= 'a' && k <= 'f'))
					{
						if (k <= '9') k -= '0';
						else if (k <= 'F') k -= 'A' - 10;
						else k -= 'a' - 10;
						if (low)
						{
							data -> SetDato (curY,(data -> GetDato (curY) & 0xf0) | (k & 0x0f));
							low = false;
							data -> saveValue (curY);
							if (curY < data -> getSize ()) curY ++;
						}
						else
						{
							data -> SetDato (curY, (data -> GetDato (curY) & 0x0f) | ((k << 4) & 0xf0));
							low = true;
							data -> saveValue (curY);
						}
					}
					break;
				case 1 :        // Char
					data -> SetDato (curY, k);
					data -> saveValue (curY);
					if (curY < data -> getSize ()) curY ++;
					break;
				case 2 :        // Asm
					ChangeIst ();
					if (CIstLen == 0 || curX >= EDLEN) break;

					if (k >= ' ' && k < 127)
					{
						EdBuf [curX++] = k;
						if (curX > CIstLen)
						{
							CIstLen = curX;
							EdBuf [curX] = 0;
						}
					}
					break;
			}
			queue_draw_area (redraw_rect.x, redraw_rect.y, redraw_rect.width, redraw_rect.height);
      		break;
	}
  	// Visualizza la posizione del cursore
	if (curY < yOffs) {
		VAdj->set_value (curY);
	} else if (curY > yOffs + nLines - 3) {
		if (curY < yOffs + 2 * nLines - 3) {
			// Se la nuova posizione sta dentro le line calcolate
			// Si posizione su di una linea di inizio istruzione
			nextOffs = (Ists[curY - nLines - yOffs + 2][0]) ? curY - nLines + 2 : 
				Next [curY - nLines - yOffs + 2] - data -> getStart ();
	    	VAdj->set_value (nextOffs);
		} else {
	    	VAdj->set_value (curY - nLines + 2);
		}
	} 
	if (curPos () < xOffs) {
		HAdj->set_value (curPos ());
	} else if (curPos () >= xOffs + (get_width () - border)/text_w) {
		HAdj->set_value (curPos () - (get_width () - border)/text_w + 1);
	}
}

bool CMemoryPanel::on_expose_event (GdkEventExpose *event)
{
    DisassView ();
    // Se un'area non e' stata letta, la legge
	((XDataMemory*) data) ->checkRange (yOffs + data -> getStart (), ((curH - border)/text_h + yOffs)  + data -> getStart ());
 
     return MemoryPanel::on_expose_event (event);
}

unsigned CMemoryPanel::curPos ()
{
	switch (Zone)
	{
		case 0:	// Hex
			return (low) ? 7 : 6;
		case 1:	// Ascii
			return 10;
		case 2:	// Assembler
			return 13 + curX;
	}
	return 0;
}


void CMemoryPanel::menuAddBk ()
{
	sfr -> addBkp (curY + data -> getStart ());
	RefreshLines (1);
}

void CMemoryPanel::menuDelBk ()
{
	sfr -> delBkp (curY + data -> getStart ());
	RefreshLines (1);
}


bool CMemoryPanel::on_button_press_event (GdkEventButton *b)
{
unsigned newY;
unsigned x;
	if (b -> type == GDK_2BUTTON_PRESS) {
		// Quando entra qui, ha gia' sistemato curY, quindi usa quello
		if (sfr -> isBp (curY + data -> getStart ())) {
			menuDelBk ();
		} else {
			menuAddBk ();
		}
		return true;
	}
	curOn = true;
	tick();
	newY = ((int) b -> y - border)/text_h + yOffs;
	x = ((int) b -> x - border)/text_w + xOffs;
	if (Zone == 2 &&			// Se e' in 'assembler'
		(curY != newY ||			// ha cambiato linea o
			x < ASMPOS) &&		// ha cambiato zona
				Asm ()) {			// e l'ASM e' sbagliato
		return true;
	}

	if (x < CHARPOS) {
		Zone = 0;
		low = (x < HINIBPOS) ? false : true;
	} else if (x < ASMPOS) {
		Zone = 1;
	} else {
		if (curY != newY || Zone != 2) {
			Zone = 2;
			if (Ists [newY-yOffs][0]==0) {	// Se la linea non contiene assembler
				newY = Prev [newY-yOffs] - data->getStart();
			}
			curY = newY;	// Si posiziona sulla nuova riga per mettere nel buffer la cosa giusta
			ChangeIst ();
		}
		curX = x - ASMPOS;
		if (curX > CIstLen) curX = CIstLen;
	}
	curY = newY;

	if ( b -> button == 3) {
	int PC = curY + data -> getStart ();
		if (sfr -> isBp (PC) && popupMenuReset) {
			popupMenuReset->popup(b->button, b->time);
		} else if (sfr -> hasBp () && sfr -> checkBps (PC) && popupMenuSet) {
			popupMenuSet->popup(b->button, b->time);
		} 
	} 
	return true;
}

void CMemoryPanel::RefreshLines (int n)
{
	queue_draw_area (0, (curY - yOffs) * text_h + border, get_width (), text_h * n);
}

void CMemoryPanel::DisassView (void)
{
int start = data -> getStart ();
int pc = yOffs + start;
int n = (get_height () - 2 * border) / text_h + 1;
int i,k;
	Prev [0] = (pc > start) ? pc - 1 : start;
	for (i = 0; i < n; i++ )
	{       // for each line:
    int j = yOffs + i + start;
		Ists [i][0] = 0;        // nessuna istruzione
		if (j == pc)
		{       // Linea che contiene un'istruzione
			pc += processor -> GetMnemonic (pc, Ists [i]);
			for (k = j;k < pc;k ++) {
				Next [k - yOffs - start] = pc;
			}
			for (k = j + 1;k <= pc;k ++) {
				Prev [k - yOffs - start] = j;
			}
		}
	}
}

bool CMemoryPanel::Asm (void)
{
int j;
	curX = 0;
	if (!IstChanged) {
		return false;
	}
	j = processor -> Asm (EdBuf,curY+data->getStart());
	if (!j) {
		return true;	// Errore
	}
	IstChanged = false;
	DisassView ();	// Ricalcola gli offset
	RefreshLines (j);
	return false;	// OK
}

bool CMemoryPanel::on_configure_event (GdkEventConfigure *event)
{
bool res = MemoryPanel::on_configure_event (event);
	// Ridimensiona gli array per l'editing del codice
	if (!Ists || nLines != (get_height () -  border)/text_h + 1) {
	unsigned  i;
		if (Ists) {
			delete Ists;
		}
		if (Next) {
			delete Next;
		}
		if (Prev) {
			delete Prev;
		}
		nLines = (get_height () - border)/text_h + 1;
		Ists = new char [nLines][EDLEN];
		for (i = 0;i < nLines; i ++) {
			Ists [i][0] = 0;	// Svuota le linee, per evitare errori
		}
		Next = new unsigned [nLines + 3];
		Prev = new unsigned [nLines + 3];
	}
	return res;
}

void CMemoryPanel::goTo (unsigned where) {
	position (where, true);
}

CMemoryPanel::CMemoryPanel (i8052 *proc, Pcmon *p) : 
    MemoryPanel (proc->GetCode (), p),
	popupMenuSet (0),
	popupMenuReset (0)
{
  sfr = proc->GetSFR();
  IstChanged = 0;
  Ists = NULL;
  Next = Prev = NULL;
  processor = proc;
	
  //Create actions for menus and toolbars:
  m_refActionGroup = Gtk::ActionGroup::create();

  m_refActionGroup->add( Gtk::Action::create("SetBrk", "Imposta Breackpoint"),
  	sigc::mem_fun(*this, &CMemoryPanel::menuAddBk) );
  m_refActionGroup->add( Gtk::Action::create("ResetBrk", "Elimina Breackpoint"),
  	sigc::mem_fun(*this, &CMemoryPanel::menuDelBk) );
  m_refActionGroup->add( Gtk::Action::create("EseguiA", "Esegui Fino a Questo Punto"),
  	sigc::mem_fun(*this, &CMemoryPanel::menuRunTo) );
  m_refActionGroup->add( Gtk::Action::create("EseguiDa", "Esegui Da Questo Punto"),
  	sigc::mem_fun(*parent, &Pcmon::menuRunPc) );

  m_refUIManager = Gtk::UIManager::create();
  m_refUIManager->insert_action_group(m_refActionGroup);
  // add_accel_group(m_refUIManager->get_accel_group());

  try
  {
    /* Crea l'immagine per tutti i menu (principale e popup) e le toolbar */
    Glib::ustring ui_info =
        "<ui>"
        "  <popup name='PopupMenuSet'>"
        "    <menuitem action='SetBrk'/>"
        "    <menuitem action='EseguiA'/>"
        "    <menuitem action='EseguiDa'/>"
        "  </popup>"
        "  <popup name='PopupMenuReset'>"
        "    <menuitem action='ResetBrk'/>"
        "    <menuitem action='EseguiA'/>"
        "    <menuitem action='EseguiDa'/>"
        "  </popup>"
        "</ui>";

    m_refUIManager->add_ui_from_string(ui_info);
  }
  catch(const Glib::Error& ex)
  {
    std::cerr << "Preparazione dei menu fallita: " <<  ex.what();
  }

  //Get the toolbar widgets, and add them to a container widget:
  popupMenuSet = dynamic_cast<Gtk::Menu*>( m_refUIManager->get_widget("/PopupMenuSet") ); 
  popupMenuReset = dynamic_cast<Gtk::Menu*>( m_refUIManager->get_widget("/PopupMenuReset") ); 
  if(!popupMenuSet || !popupMenuReset)
    g_warning("Manca un menu popup della finestra del codice");
}
