/***************************************************************************
 *  dmemory.cc
 *
 *  Sun Nov 14 10:49:20 2004
 *  Copyright  2004 Ing. Stefano Salvi
 *  Email stefano@salvi.mn.it
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
/*
 * DataMemory
 * Specializza per memoria dati
 */
#include <stdio.h>

#include "dmemory.h"
#include "sfrnames.h"
#include "serial.h"

int HGPRegister::getValue ()
{
	return mem->GetRegister(regnum); 
}

void HGPRegister::setValue (int v) // Modifica il valore e lo scrive sulla scheda
{ 
	mem->SetRegister(regnum,v); 
	mem -> r -> redraw ();
	store (); 
}

HRegister &HGPRegister::operator = (int i)						// Modifica il valore ed aggiorna campo associato
{
	mem->SetRegister(regnum, i & 0xff);	// Limita a 8/16 bit
	redraw ();
	mem -> redraw ();
	return *this;
}

void HGPRegister::store ()
{
char buf [20];
	Serial_Container::s -> schPutc (set [0]);	// Comando
	if (Serial_Container::s -> schWaitc (' ')) {
		return;
	}
	Serial_Container::s -> schPutc (set [1]);  // Registro
	if (Serial_Container::s -> schWaitc (' ')) {
		return;
	}
	sprintf (buf,"%02x", mem->GetRegister(regnum));
	Serial_Container::s -> schPuts (buf);
	Serial_Container::s -> schPrompt ();
}

/*
 * DataMemory
 */

DataMemory::DataMemory (int c_start, int c_size, SFRMemory *c_SFR) : 
	Memory (c_start, c_size)
{
	Regs[0]= new HGPRegister(0, "C0", this);
	Regs[1]= new HGPRegister(1, "C1", this);
	Regs[2]= new HGPRegister(2, "C2", this);
	Regs[3]= new HGPRegister(3, "C3", this);
	Regs[4]= new HGPRegister(4, "C4", this);
	Regs[5]= new HGPRegister(5, "C5", this);
	Regs[6]= new HGPRegister(6, "C6", this);
	Regs[7]= new HGPRegister(7, "C7", this);
	SFR = c_SFR; 
}

unsigned DataMemory::GetDPTR (void)
{
	return SFR -> GetDPTR () & 0xffff;
}

void DataMemory::saveValue (int pos)
{
char buf [10];
	Serial_Container::s->schPutc ('A');
	if (Serial_Container::s -> schWaitc (' ')) {
		return;
	}
	sprintf (buf,"%02X", pos);
	Serial_Container::s -> schPuts (buf);
	if (Serial_Container::s -> schWaitc ('-')) {
		return;
	}
	sprintf (buf,"%02X\r", data [pos]);
	Serial_Container::s -> schPuts (buf);
	Serial_Container::s -> schPrompt ();
}

static int BitMsks [] = { 1,2,4,8,16,32,64,128 };      // Mschere dei bit

void DataMemory::SetDDato (unsigned ad,char val)
{
	if (ad < 128) {
		SetDato  (ad,val);
	} else {
		SFR -> SetSFR (ad,val & 0xff);
	}
}

unsigned char DataMemory::GetDDato (unsigned ad)
{
	return (ad < 128) ? GetDato (ad) : SFR -> GetSFR (ad);
};

int DataMemory::GetBit (int Bit)
{
int Word;
	if (Bit < 128) {
		Word = data [(Bit >> 3) + 0x20];
	} else {
		Word = SFR -> GetSFR (Bit & 0xf8);
	}
	return Word & BitMsks [Bit & 0x7];
}

void DataMemory::SetBit (int Bit,int val)
{
int Word;
	if (Bit < 128) {
		Word = data [(Bit >> 3) + 0x20];
	} else {
		Word = SFR -> GetSFR (Bit & 0xf8);
	}
	if (val) {
		Word |= BitMsks [Bit & 0x7];
	} else {
		Word &= ~BitMsks [Bit & 0x7];
	}
	if (Bit < 128) {
		data [(Bit >> 3) + 0x20] = Word;
	} else {
		SFR -> SetSFR (Bit & 0xf8,Word);
	}
}

unsigned DataMemory::GetTopOfStack (void)
{
int Sp = SFR -> GetSP () & 0xff;
	return ((data [Sp] << 8) & 0xff00) | (data [Sp - 1] & 0xff);
}

unsigned char DataMemory::GetRegister (int Reg)
{
  unsigned char *base = data + getRB ();
  return base [Reg & 0x07]; 
}

void DataMemory::SetRegister (int Reg,unsigned char val)
{
unsigned char *base = data + getRB ();
	base [Reg & 0x07] = val;
	Regs [Reg & 0x07] -> redraw ();
}

void DataMemory::Push (char v)
{
int stack;
	stack = (GetDDato (aSP) + 1) & 0xff;
	SetDato (stack,v);
	SetDDato (aSP,stack);
}

char DataMemory::Pop (void)
{
int stack;
	stack = (GetDDato (aSP)) & 0xff;
	SetDDato (aSP,stack - 1);
	return GetDato (stack);
}

void DataMemory::SFRreload ()
{
char bpbuf [60];
char buf [80];
char buf1 [80];
int localPC, localB0, localB1, localB2, localB3;
int SP, A, B, PSW, DPTR;
int R [8];
int i;

	Serial_Container::s -> schPuts ("B ");
	if (!Serial_Container::s -> schGets (bpbuf,59) &&
		!Serial_Container::s -> schWaitc ('*')) {
		Serial_Container::s -> schPutc ('R');
		if (!Serial_Container::s -> schGets (buf,79) &&
			!Serial_Container::s -> schGets (buf1,79)) {
			Serial_Container::s -> schPrompt ();
			sscanf (bpbuf, "B B0=%x B1=%x B2=%x B3=%x" , &localB0, &localB1, &localB2, &localB3);
			SFR->SetB0 (0);		// Cancella sempre incondizionatamente il breackpoint 0
			SFR->SetB1 (localB1);
			SFR->SetB2 (localB2);
			SFR->SetB3 (localB3); 

			sscanf (buf, "R PC=%x SP=%x A=%x B=%x PSW=%x DPTR=%x", &localPC, &SP, &A, &B, &PSW, &DPTR);
			SFR->SetPc (localPC);
			sscanf (buf1, "R0=%x R1=%x R2=%x R3=%x R4=%x R5=%x R6=%x R7=%x" , R, R + 1, R + 2, R + 3,
				R + 4, R + 5, R + 6, R + 7);
			SFR->SetSP (SP);
			SFR->SetAcc (A);
			SFR->SetBreg (B);
			SFR->SetPSW (PSW);
			SFR->SetDPTR (DPTR);
			for (i = 0; i < 8; i++) {
				*Regs [i] = R [i];
			}
			return;
		}
	}
	// Offline (emulatore)
	if (SFR->GetPc() < 0x8050) {
		SFR->SetPc (0x8050);
	}
	SFR -> redraw ();
	for (i=0; i < 8; i ++) {
		Regs [i]->redraw ();
	}
	r->redraw ();
}

void DataMemory::redraw ()
{
int i;
	Memory::redraw ();
	SFR -> redraw ();
	for (i=0; i < 8; i ++) {
		Regs [i]->redraw ();
	}
}

void DataMemory::setStatus (stat_val s)
{
int i;
	Memory::setStatus (s);
	for (i=0; i < 8; i ++) {
		Regs [i]->setStatus (s);
	}
}

void DataMemory::redrawIfChanged ()
{
int i;
	Memory::redrawIfChanged ();
	SFR -> redrawIfChanged ();
	for (i=0; i < 8; i ++) {
		Regs [i]->redrawIfChanged ();
	}
}

/*
 * Ricarica il contenuto completo della memoria
 */
void DataMemory::reload ()
{
int i,j;
char inbuf [50];
int vals [8];

	Serial_Container::s->schPutc ('I');
	if (Serial_Container::s->schWaitc (' ')) {
		return;
	}
	Serial_Container::s->schPuts ("00");
	if (Serial_Container::s->schWaitc (',')) {
		return;
	}
	Serial_Container::s->schPuts ("FF");
	if (Serial_Container::s->schGets (inbuf,49)) {	// Risucchia linea di comando
		return;
	}
	for (i = 0; i < 256; i += 8)
	{
		if (Serial_Container::s -> schGets (inbuf,49)) {
			return;	// Scheda non collegata
		}
		sscanf (inbuf,"%02x  %02x %02x %02x %02x %02x %02x %02x %02x", &j,
			vals, vals + 1, vals + 2, vals + 3, vals + 4, vals + 5, vals + 6, vals + 7);
		for (j = 0; j < 8; j++) {
			SetDato (i + j, vals [j]);
		}
	}
	Serial_Container::s -> schPrompt ();

	r -> redraw ();
}
