/***************************************************************************
 *  editwindow.cc
 *
 *  Sun Nov 14 10:49:20 2004
 *  Copyright  2004 Ing. Stefano Salvi
 *  Email stefano@salvi.mn.it
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <string.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <regex.h>


#include <glib/gmacros.h>
#include <glib.h>
#include <gtkmm.h>
#include <libgnomemm.h>

#include <iostream>	// Per la stampa dell'eventuale errore

#include "editbuffer.h"
#include "editwindow.h"
#include "messageinterface.h"
#include "dialogs.h"
#include "finddialog.h"
#include "pcmon.h"

FindDialog *EditWindow::findDialog = NULL;

// #define 	EXTERNAL_ASSEMBLER

/*
 * myEdit
 */

myEdit::myEdit (EditWindow *p)
{ 
	parent=p;
}

bool myEdit::on_button_press_event (GdkEventButton *event) {
Glib::RefPtr<Gtk::TextBuffer> cx = get_buffer ();
bool status;
	if ( event->button == 3) {
    	if(parent->m_pMenuPopup)
      		parent->m_pMenuPopup->popup(event->button, event->time);

		return true;
	}
	status =  Gtk::TextView::on_button_press_event (event);
	parent->setMessage ();	// Aggiorna le coordinate nella barra di stato
	return status;
}

/*
 * EditWindow
 */
void EditWindow::menuRedo ()
{
	cx -> undo_redo_common(EditBuffer::c_redo);
	editbox.scroll_mark_onscreen (cx -> get_insert ());
	setMessage ();	// Aggiorna le coordinate nella barra di stato
}

void EditWindow::menuUndo ()
{
	cx -> undo_redo_common (EditBuffer::c_undo);
	editbox.scroll_mark_onscreen (cx -> get_insert ());
	setMessage ();	// Aggiorna le coordinate nella barra di stato
}

void EditWindow::menuUndoAll ()
{
	cx -> undoAll ();
	editbox.scroll_mark_onscreen (cx -> get_insert ());
	setMessage ();	// Aggiorna le coordinate nella barra di stato
}

void EditWindow::menuRedoAll ()
{
	cx -> redoAll ();
	editbox.scroll_mark_onscreen (cx -> get_insert ());
	setMessage ();	// Aggiorna le coordinate nella barra di stato
}

void EditWindow::menuGoTo ()
{
int linenum;
char *str;
	str = LabelDialog::doGetText("Vai alla riga",  "Vai");
	if (str) {
		linenum = strtol (str,NULL,0);
		lineSelect (linenum, true);
  		delete (str);
  	}
}

void EditWindow::menuAndToolbarBuild ()
{
  //Create actions for menus and toolbars:
  m_refActionGroup = Gtk::ActionGroup::create();

  m_refActionGroup->add( Gtk::Action::create("MenuFile", "_File") );
  m_refActionGroup->add( Gtk::Action::create("Nuovo", Gtk::Stock::NEW),
  	sigc::mem_fun(*parent,&Pcmon::menuNew) );
  m_refActionGroup->add( Gtk::Action::create("Apri", Gtk::Stock::OPEN),
  	sigc::mem_fun(*parent,&Pcmon::menuOpen) );
  m_refActionGroup->add( Gtk::Action::create("Salva", Gtk::Stock::SAVE),
  	sigc::mem_fun(*this,&EditWindow::menuSave) );
  m_refActionGroup->add( Gtk::Action::create("SalvaCome", Gtk::Stock::SAVE_AS),
  	sigc::mem_fun(*this,&EditWindow::menuSaveAs) );
  m_refActionGroup->add( Gtk::Action::create("Chiudi", Gtk::Stock::CLOSE),
  	sigc::mem_fun(*this,&EditWindow::menuClose) );
  m_refActionGroup->add( Gtk::Action::create("Fine", Gtk::Stock::QUIT),
  	sigc::mem_fun(*parent,&Pcmon::menuQuit) );
	
  m_refActionGroup->add( Gtk::Action::create("MenuEdit", "_Modifica") );
  m_refActionGroup->add( Gtk::Action::create("Taglia",  Gtk::Stock::CUT),
  	sigc::mem_fun(*this,&EditWindow::menuCut) );
  m_refActionGroup->add( Gtk::Action::create("Copia", Gtk::Stock::COPY),
  	sigc::mem_fun(*this,&EditWindow::menuCopy) );
  m_refActionGroup->add( Gtk::Action::create("Incolla", Gtk::Stock::PASTE),
  	sigc::mem_fun(*this,&EditWindow::menuPaste) );
  m_refActionGroup->add( Gtk::Action::create("Cancella", Gtk::Stock::DELETE),
  	sigc::mem_fun(*this,&EditWindow::menuDelete) );
  m_refActionGroup->add( Gtk::Action::create("SelezionaTutto",  "Seleziona Tutto"),
  	sigc::mem_fun(*this,&EditWindow::menuSelAll) );
  undoButton = Gtk::Action::create("Annulla", Gtk::Stock::UNDO);
  m_refActionGroup->add( undoButton,
  	sigc::mem_fun(*this,&EditWindow::menuUndo) );
  undoButton -> set_sensitive (false);
  redoButton = Gtk::Action::create("Ripristina", Gtk::Stock::REDO);
  m_refActionGroup->add( redoButton,
  	sigc::mem_fun(*this,&EditWindow::menuRedo) );
  redoButton -> set_sensitive (false);
	
  m_refActionGroup->add( Gtk::Action::create("AnnullaTutto", "Annulla Tutto"),
  	sigc::mem_fun(*this,&EditWindow::menuUndoAll) );
  m_refActionGroup->add( Gtk::Action::create("RipristinaTutto",  "Ripristina Tutto"),
  	sigc::mem_fun(*this,&EditWindow::menuRedoAll) );
  m_refActionGroup->add( Gtk::Action::create("Cerca", Gtk::Stock::FIND_AND_REPLACE),
  	sigc::mem_fun(*this,&EditWindow::menuFind) );
  m_refActionGroup->add( Gtk::Action::create("CercaAncora", Gtk::StockID("pcmon_btn_findnext")),
  	sigc::mem_fun(*this,&EditWindow::menuFindAgain) );
  m_refActionGroup->add( Gtk::Action::create("VaiA", Gtk::Stock::JUMP_TO),
  	sigc::mem_fun(*this,&EditWindow::menuGoTo) );

  m_refActionGroup->add( Gtk::Action::create("MenuProgramma", "_Programma") );
  
  m_refActionGroup->add( Gtk::Action::create("Assembla", Gtk::StockID("pcmon_btn_assemble")),
  	sigc::mem_fun(*this,&EditWindow::menuAssemble) );
  m_refActionGroup->add( Gtk::Action::create("Scarica", Gtk::StockID("pcmon_btn_upload")),
  	sigc::mem_fun(*this,&EditWindow::menuUpload) );
  m_refActionGroup->add( Gtk::Action::create("Esegui", Gtk::Stock::EXECUTE),
  	sigc::mem_fun(*this,&EditWindow::menuRun) );

  m_refUIManager = Gtk::UIManager::create();
  m_refUIManager->insert_action_group(m_refActionGroup);
  add_accel_group(m_refUIManager->get_accel_group());

  try
  {
    /* Crea l'immagine per tutti i menu (principale e popup) e le toolbar */
    Glib::ustring ui_info =
        "<ui>"
        "  <menubar name='MenuBar'>"
        "    <menu action='MenuFile'>"
        "      <menuitem action='Nuovo'/>"
        "      <menuitem action='Apri'/>"
        "      <menuitem action='Salva'/>"
        "      <menuitem action='SalvaCome'/>"
        "      <menuitem action='Chiudi'/>"
        "      <separator/>"
        "      <menuitem action='Fine'/>"
        "    </menu>"
        "    <menu action='MenuEdit'>"
        "      <menuitem action='Taglia'/>"
        "      <menuitem action='Copia'/>"
        "      <menuitem action='Incolla'/>"
        "      <menuitem action='Cancella'/>"
        "      <menuitem action='SelezionaTutto'/>"
        "      <separator/>"
        "      <menuitem action='Annulla'/>"
        "      <menuitem action='Ripristina'/>"
        "      <menuitem action='AnnullaTutto'/>"
        "      <menuitem action='RipristinaTutto'/>"
        "      <separator/>"
        "      <menuitem action='Cerca'/>"
        "      <menuitem action='CercaAncora'/>"
        "      <menuitem action='VaiA'/>"
        "    </menu>"
        "    <menu action='MenuProgramma'>"
        "      <menuitem action='Assembla'/>"
        "      <menuitem action='Scarica'/>"
        "      <menuitem action='Esegui'/>"
	  	"	 </menu>"
        "  </menubar>"
        "  <toolbar  name='ToolBar'>"
        "    <toolitem action='Nuovo'/>"
        "    <toolitem action='Apri'/>"
        "    <toolitem action='Salva'/>"
        "    <toolitem action='Chiudi'/>"
        "    <separator/>"
        "    <toolitem action='Taglia'/>"
        "    <toolitem action='Copia'/>"
        "    <toolitem action='Incolla'/>"
        "    <toolitem action='Cancella'/>"
        "    <separator/>"
        "    <toolitem action='Annulla'/>"
		"    <toolitem action='Ripristina'/>"
        "    <separator/>"
        "    <toolitem action='Cerca'/>"
		"    <toolitem action='CercaAncora'/>"
		"    <toolitem action='VaiA'/>"
        "    <separator/>"
        "    <toolitem action='Assembla'/>"
		"    <toolitem action='Scarica'/>"
		"    <toolitem action='Esegui'/>"
        "  </toolbar>"
        "  <popup name='PopupMenu'>"
        "    <menuitem action='Taglia'/>"
        "    <menuitem action='Copia'/>"
        "    <menuitem action='Incolla'/>"
        "    <menuitem action='Cancella'/>"
        "  </popup>"
        "</ui>";

    m_refUIManager->add_ui_from_string(ui_info);
  }
  catch(const Glib::Error& ex)
  {
    std::cerr << "Preparazione dei menu fallita: " <<  ex.what();
  }

  
  //Get the menubar and toolbar widgets, and add them to a container widget:
  Gtk::Widget* pMenubar = m_refUIManager->get_widget("/MenuBar");
  if(pMenubar)
  {
    vbox.pack_start(*pMenubar, Gtk::PACK_SHRINK);
	pMenubar->signal_button_press_event().connect(sigc::mem_fun(*this,&EditWindow::mouseDown));
  }
  Gtk::Toolbar* toolbar = (Gtk::Toolbar *) m_refUIManager->get_widget("/ToolBar") ;
  if(toolbar)
    vbox.pack_start(*toolbar, Gtk::PACK_SHRINK);
  toolbar->set_toolbar_style(Gtk::TOOLBAR_ICONS);
  
  m_pMenuPopup = dynamic_cast<Gtk::Menu*>( m_refUIManager->get_widget("/PopupMenu") ); 
  if(!m_pMenuPopup)
    g_warning("menu not found");

}

bool EditWindow::on_event(GdkEvent* p1)
{
	if (((GdkEventAny*) p1) -> type == GDK_KEY_RELEASE) {
	  setMessage ();	// Aggiorna le coordinate nella barra di stato
	}
	if (((GdkEventAny*) p1) -> type == GDK_FOCUS_CHANGE && ((GdkEventFocus*) p1) -> in) {
    	// Registra la nuova finestra attiva nel dialog di ricerca
		if (findDialog) {
			findDialog -> setFrontWin (cx);
		}
	}
	return false;
}

char *EditWindow::newChars (std::string const &str)
{
int l = str.length ();
char *r = new char [l + 1];
	strncpy (r,str.data (),l);
	r [l] = 0;
	return r;
}

void EditWindow::menuUpload ()
{
	doUpload (NULL);
}

void EditWindow::menuRun ()
{
char *file = objName ();
	if (testChanged (file) || !uploaded) {
		if (doUpload (file)) {
      		return;
		}
	} else {
		delete file;
	}
	parent -> menuRunStart ();
}

char *EditWindow::objName ()
{
	if (cx->getFilename()) {
	char *file = new char [strlen (cx->getFilename()) + 5];
	char *s,*s1;
		strcpy (file,cx->getFilename());
		s = strrchr (file,'.');
		s1 = strrchr (file,'/');
		if (!s || s1 > s) {
			s = file + strlen (file);
		}
		strcpy (s,HEXEXTENSION);
		return file;
	}
	return NULL;
}

int EditWindow::testChanged (char *file)
{
struct stat sobj;
struct stat ssrc;
	if (cx -> isChanged ()) {
		return true;
	} else if (stat (file,&sobj)) {
		return true;	// Non esiste l'obj
	} else if (cx->getFilename()) {
		stat (cx->getFilename(),&ssrc);
		return sobj.st_mtime < ssrc.st_mtime || sobj.st_ctime < ssrc.st_ctime;
	}
	return false;
}

int EditWindow::doUpload (char *file)
{
	if (!file) {
		file = objName ();
		if (!file) {
			return true;
		}
	}
	if (testChanged (file) && doAssemble ()) {
		delete file;
		return true;
	}
	parent -> doUpload (file);
	uploaded = true;
	delete file;
	return false;
}

void EditWindow::menuAssemble ()
{
	doAssemble ();
}

/* Lancia l'assembler, raccoglie gli errori eventuali in una lista
 * Ritorna:	false se ci sono errori
 *		true se e' andato tutto bene
 */
bool EditWindow::doAssemble ()
{
int err = false;
#ifdef EXTERNAL_ASSEMBLER
#define MSGLEN	128
int myPipe [2];
char msg [MSGLEN];
int pid;
int status;

	if (cx -> isChanged () || !cx->getFilename()) {
		if (!realSave (true)) {
			return false;
		}
	}

	uploaded = false;			// Non e' piu' 'scaricato', sto modificandolo
	// Crea la pipe per catturare i risultati
	if (pipe (myPipe) < 0) {
		perror ("myPipe");
	}

	// crea il sottoprocesso
	pid = fork ();
	if (pid < 0) {	// Errore di creazione del sottoptocesso
		perror ("Fork");
	} else if (pid) {	// Parent
		// Costruisce un 'FILE' su un file di basso livello
		FILE *assembler = fdopen (myPipe [0],"r");
		if (!assembler) {
			perror ("File Assembler");
			return true;
		}
	
		::close (myPipe [1]);	// Chiude la sua copia parte sbagliata della pipe
	
		setbuf (assembler,0);

		// Se c'era gia' il dialog con gli errori, lo svouta
		if (errorDialog) {
			errorDialog -> ClearErrors ();
		}
		// Attende che il processo figlio chiuda il file (la pipe)
		while (!feof (assembler))
		{
			if (fgets (msg,MSGLEN,assembler)) { // Legge la stringa stampata dall'asseblatore nella pipe
				if (strstr (msg,"line")) {  // Se contiene la parola chiave 'line', e' un errore
					AddError (msg);  // Aggiunge l'errore al dialog
				}
			}
		}
		waitpid (pid,&status,0);
		if (!(errorDialog && errorDialog->hasErrors()) && WEXITSTATUS(status)) {
			sprintf (msg,"Warning, Line 0, L'assemblatore ha ritornato il codice di errore %d",
				WEXITSTATUS(status));
			AddError (msg);  // Aggiunge l'errore al dialog
		}
    	if (errorDialog && errorDialog->hasErrors()) {	// L'assembler ha dato errori
			errorDialog -> display ();
			err = true;
		}
	} else {	// child
		// Redirige l'output (out ed err) sulla pipe
		// dato che sono gia' nel 'child', non devo fare copie
		dup2 (myPipe [1],STDERR_FILENO);
		dup2 (myPipe [1],STDOUT_FILENO);
		// Chiude la pipe originaria
		::close (myPipe [1]);
	
		::close (myPipe [0]);	// Chiude la sua copia parte sbagliata della pipe

		if (execlp ("as31","-Fhex","-l",cx->getFilename(),NULL)) {
			perror ("Esecuzione");
			exit (0);
		}
	}
	::close (myPipe [0]);	// chiude anche il mio lato della pipe
#else
	if (cx -> isChanged () || !cx->getFilename()) {
		if (!realSave (true)) {
			return true;
		}
	}

	uploaded = false;			// Non e' piu' 'scaricato', sto modificandolo

	if (errorDialog) {
		errorDialog -> ClearErrors ();
	}
	
	parent->assembler (cx->getFilename(), this);
	
    if (errorDialog && errorDialog->hasErrors()) {	// L'assembler ha dato errori
		errorDialog -> display ();
		err = true;
	}
#endif
	return err; 
}

void EditWindow::AddError (char *msg)
{
	if (!errorDialog) {  // Se il dialog di errore non esiste gia', lo crea
		errorDialog = new ErrorDialog (this);
	}
	errorDialog -> AddError (msg);  // Aggiunge l'errore al dialog
}

/* Salva il file, chede il nome se non ne ha gia uno
 * Ritorna:	se ha chiesto il nome ed e' stato premuto "annulla"
 *		true se e' andato tutto bene
 */
bool EditWindow::realSave (bool force)
{
	if (cx -> isChanged () || (!cx->getFilename() && force)) {
		if (!cx->getFilename()) {
			return realSaveAs ();
		} else {
			cx -> file_save_execute(cx->getFilename());
			set_title (get_title ());
		}
	}
	return true;
}

/* Salva il file, chede il nome se non ne ha gia uno
 * Ritorna:	se ha chiesto il nome ed e' stato premuto "annulla"
 *		true se e' andato tutto bene
 */
bool EditWindow::realSaveAs ()
{
int result;
	Gtk::FileSelection dialog("Indica il nome con cui salvzare il sorgente");
	dialog.set_transient_for(*this);

	result = dialog.run();

	//Handle the response:
	if (result == Gtk::RESPONSE_OK) {
		int len = dialog.get_filename().length ();
		if (len) {
			struct stat stmp;
			char *file = new char [len + 1];
			strncpy (file, dialog.get_filename().data (),len);
			file [len] = 0;

			if (!stat (file,&stmp)) { // Il file esiste
				if (MessageInterface::doYesNoDialog ("Salva con nome", "Sovrascrivo", "Annullo", 
					"Il file '%s' esiste", file) == GTK_RESPONSE_ACCEPT)
				delete (file);
      			return false;
			}

			if (cx -> file_save_execute (file)) {
				cx -> setFilename (file);
				set_title (cx -> get_title ());
			} else {
				delete (file);
			}
		}
		return true;
	}
	return false;
}

void EditWindow::menuCut ()
{
	cx->cut_clipboard(Gtk::Clipboard::get(GDK_SELECTION_CLIPBOARD));
	setMessage ();	// Aggiorna le coordinate nella barra di stato
}

void EditWindow::menuCopy ()
{
	cx->copy_clipboard(Gtk::Clipboard::get(GDK_SELECTION_CLIPBOARD));
}

void EditWindow::menuPaste ()
{
	cx->paste_clipboard(Gtk::Clipboard::get(GDK_SELECTION_CLIPBOARD));
	setMessage ();	// Aggiorna le coordinate nella barra di stato
}

void EditWindow::menuDelete ()
{
	cx->erase_selection(true, true);
	setMessage ();	// Aggiorna le coordinate nella barra di stato
}

void EditWindow::menuSelAll ()
{
	cx->place_cursor (cx -> begin ());
	cx->move_mark (cx-> get_selection_bound (), cx -> end());
	setMessage ();	// Aggiorna le coordinate nella barra di stato
}

void EditWindow::menuFind ()
{
	if (!findDialog) {
		findDialog = new FindDialog ();
	}
	findDialog -> doFind (cx);
}

void EditWindow::menuFindAgain ()
{
	if (findDialog) {
		findDialog -> reDoFind (cx);
		setMessage ();	// Aggiorna le coordinate nella barra di stato
	}
}

bool EditWindow::on_delete_event (GdkEventAny *e)
{
	return !realClose (true);
}

void EditWindow::menuClose ()
{
    realClose (true);
}

int EditWindow::realClose (int single)
{
	if (cx -> isChanged ()) {
		int res = MessageInterface::doYesNoCancelDialog ("Chiusura", "Salvo", "Non Salvo", "Annullo", 
			"Il file '%s' e' stato modificato", cx->getFilename());
		if (res == GTK_RESPONSE_ACCEPT) {	// Salva
			if (!realSave (false)) {
				return false;
			}
		} else if (res == GTK_RESPONSE_CANCEL) {	// Annulla
			return false;
		}
	}

	if (errorDialog) {
		errorDialog -> closeErrors (this);
	}
	hide ();
	/* Siccome non riesco ad intercettare l'evento 'destroy', ma la
	 * distruzione dell'oggetto avviene tramite questa funzione, pongo
	 * qui la liberazione delle risorse */
	cx -> setFilename (NULL);
	cx -> emptyUndoRedo ();

	/* Elimina il testo nell'edit */
	cx->erase (cx -> begin (), cx -> end());

	if (single) {
		parent->delEditor (this);
	}
	return true;
}

void EditWindow::openfile (char *file)
{
	if (file && *file) {
		cx -> file_open_execute (file);
    }
}

bool EditWindow::mouseDown (GdkEventButton *b)
{
	parent -> windowMenu ();
	return false;
}


bool EditWindow::on_destroy_event (GdkEventAny *e)
{
	printf ("Destroy\n");
	return false;
}

int EditWindow::isChanged ()
{
	return cx -> isChanged (); 
}

void EditWindow::setMessage ()
{
char string [35];
Gtk::TextIter cursor = cx -> get_insert () -> get_iter ();
	sprintf (string, "Linea %d Colonna %d%s", cursor.get_line () + 1, cursor.get_line_offset () + 1,
		cx->isChanged () ? " [modificato]" : "");
	Message.set_text (string);
}

void EditWindow::lineSelect (int linenum,bool activate)
{
int numlines;
	numlines = cx -> get_line_count ();
	if (--linenum >= numlines) {
		linenum = numlines - 1;
	}
	cx->place_cursor(cx->get_iter_at_line (linenum));
	editbox.scroll_mark_onscreen (cx -> get_insert ());
	if (activate) {
		present ();
		editbox.grab_focus ();
	}
	setMessage ();	// Aggiorna le coordinate nella barra di stato
}

void EditWindow::goTo (int offset)
{
Gtk::TextIter cursor = cx->get_iter_at_offset (offset);
	cx->place_cursor(cursor);
	editbox.scroll_mark_onscreen (cx -> get_insert ());
	present ();
	editbox.grab_focus ();
	setMessage ();	// Aggiorna le coordinate nella barra di stato
}

EditWindow::~EditWindow () {
	if (errorDialog) {
		delete errorDialog;
	}
}

EditWindow::EditWindow (Pcmon *Parent, EditWindow *Next) :
  editbox (this),
  vbox (false, 2),
  status (false, 2),
  Message ("")
{
  errorDialog = NULL;
  next = Next;
  parent = Parent;
  uploaded = false;

	cx = EditBuffer::create(this);
	editbox.set_buffer (cx);
	Pango::FontDescription *font = new Pango::FontDescription ("Courier,monospace, 10");
	editbox.modify_font(*font); 
	delete font;
	
	add (vbox);

	menuAndToolbarBuild ();

	editbox.set_editable(true);
	cx->place_cursor (cx -> get_iter_at_offset (0));
	editbox.set_wrap_mode(Gtk::WRAP_NONE);
	set_default_size (600,400);

	vbox.pack_start(base, true, true,0);
	base.add (editbox);
	Message.set_justify (Gtk::JUSTIFY_LEFT );
    status.pack_start (Message, false, false, 0);
    vbox.pack_start (status, false, true, 0);

	set_title ("Senza titolo");
	setMessage ();

	set_events (Gdk::KEY_PRESS_MASK);
	show_all ();
}
