/***************************************************************************
 *  finddialog.cc
 *
 *  Sun Nov 14 10:49:20 2004
 *  Copyright  2004 Ing. Stefano Salvi
 *  Email stefano@salvi.mn.it
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
  
#include <glib/gmacros.h>
#include <glib.h>
#include <gtkmm.h>
#include <libgnomemm.h>

#include "dialogs.h"
#include "finddialog.h"
#include "editwindow.h"

/* ************************************************************************** *
 *                                                                            *
 * Find Dialog                                                                *
 *                                                                            *
 * ************************************************************************** */

void FindDialog::setFrontWin (Glib::RefPtr<EditBuffer> editor)
{
	parent = editor;
}

void FindDialog::doFind (Glib::RefPtr<EditBuffer> editor)
{
	parent = editor;
	show ();
	raise ();
}

bool FindDialog::on_delete_event (GdkEventAny *e)
{
	cancelButton ();
	return true;
}

void FindDialog::cancelButton ()
{
	hide ();
}

void FindDialog::FindButton ()
{
const std::string search = testo.get_string();
EditBuffer::searchFrom fFromCursor = (fromCursor.get_active ()) ? EditBuffer::fromCursor : 
  ((fromStart.get_active ()) ? EditBuffer::fromStart : EditBuffer::selectOnly);
  replace = false;
  endSearch = 0;	// Inizio ricerca -> imposta endSearch
  parent -> search_execute(&search, NULL ,fFromCursor, 
    CaseSensitive.get_active (), RegularExpression.get_active (), endSearch);
  hide ();
}

void FindDialog::reDoFind (Glib::RefPtr<EditBuffer> editor)
{
const std::string search = testo.get_string();
const std::string sReplace = toTesto.get_string();
EditBuffer::searchFrom fFromCursor = (fromCursor.get_active ()) ? EditBuffer::fromCursor : 
  ((fromStart.get_active ()) ? EditBuffer::fromStart : EditBuffer::selectOnly);
  parent -> search_execute(&search, (replace) ? &sReplace : NULL ,
    fFromCursor, CaseSensitive.get_active (), 
      RegularExpression.get_active (), endSearch);
}

void FindDialog::ReplaceButton ()
{
const std::string search = testo.get_string();
const std::string sReplace = toTesto.get_string();
EditBuffer::searchFrom fFromCursor = (fromCursor.get_active ()) ? EditBuffer::fromCursor : 
  ((fromStart.get_active ()) ? EditBuffer::fromStart : EditBuffer::selectOnly);
  replace = true;
  endSearch = 0;	// Inizio ricerca -> imposta endSearch
  parent -> search_execute(&search, &sReplace , fFromCursor, 
    CaseSensitive.get_active (), RegularExpression.get_active (), endSearch);
  hide ();
}

void FindDialog::ReplaceAll ()
{
const std::string search = testo.get_string();
const std::string sReplace = toTesto.get_string();
EditBuffer::searchFrom fFromCursor = (fromCursor.get_active ()) ? EditBuffer::fromCursor : 
  ((fromStart.get_active ()) ? EditBuffer::fromStart : EditBuffer::selectOnly);
int fCaseSensitive = CaseSensitive.get_active ();
int fRegularExpression = RegularExpression.get_active ();
  endSearch = 0;	// Inizio ricerca -> imposta endSearch
  replace = true;
  while (parent -> search_execute(&search, &sReplace , fFromCursor, 
    fCaseSensitive, fRegularExpression, endSearch))
    if (fFromCursor == EditBuffer::fromStart) 
      fFromCursor = EditBuffer::fromCursor;  // Solo la prima volta puo' essere 'dall'inizio'
  hide ();
}

FindDialog::FindDialog () :
  testo ("Cerca:"),
  fromCursor (gr, "Dalla posizione del cursore"),
  fromStart (gr, "Dall'inizio del file"),
  selectOnly (gr, "Solo nel testo selezionato"), 
  CaseSensitive ("Case Sensitive"),
  RegularExpression ("Cerca con le Espressioni Regolari"),
  toTesto ("Cambia in:"),
  Cerca ("Cerca"),
  Cambia ("Sostituisci"),
  CambiaTutti ("Sostituisci Tutti"),
  annulla ("Annulla")
{
	set_title ("Cerca o Sostituisci");
	get_vbox()->pack_start(*testo.getBox (), false,false,0);

	get_vbox()->pack_start(fromCursor, false,false,0);
	get_vbox()->pack_start(fromStart, false,false,0);
	get_vbox()->pack_start(selectOnly, false,false,0);
	fromCursor.set_active (true);
	fromStart.set_active (false);
	selectOnly.set_active (false);
	get_vbox()->pack_start(CaseSensitive, false,false,0);
	get_vbox()->pack_start(RegularExpression, false,false,0);


	get_vbox()->pack_start(*toTesto.getBox (), false,false,0);

	get_action_area()->pack_start(Cerca, false,false,0);
	get_action_area()->pack_start(Cambia, false,false,0);
	get_action_area()->pack_start(CambiaTutti, false,false,0);
	get_action_area()->pack_start(annulla, false,false,0);

	Cerca.signal_pressed() .connect(sigc::mem_fun(*this,&FindDialog::FindButton));
	Cambia.signal_pressed() .connect(sigc::mem_fun(*this,&FindDialog::ReplaceButton));
	CambiaTutti.signal_pressed() .connect(sigc::mem_fun(*this,&FindDialog::ReplaceAll));
	annulla.signal_pressed() .connect(sigc::mem_fun(*this,&FindDialog::cancelButton));

	show_all ();
}

/* ************************************************************************** *
 *                                                                            *
 * Error Dialog                                                               *
 *                                                                            *
 * ************************************************************************** */

bool ErrorDialog::on_delete_event (GdkEventAny *e)
{
	CancelButton ();
	return true;
}

void ErrorDialog::CancelButton ()
{
	hide ();
}

void ErrorDialog::GoToButton ()
{
	doSelect (true);
}

void ErrorDialog::ClearErrors ()
{
	errList -> clear ();
}

#define	LINEMSG			"line"
#define	LINEMSGLEN		4

void ErrorDialog::AddError (char *s)
{
char *t, *m;
int linnum = -1;

	for (t = s;*t >= ' '; t++);
	*t = 0;	// Elimina il cr/lf
	t = strstr (s,LINEMSG);
	if (t) {
		m = t + LINEMSGLEN;
		// Salta eventuale ':' e serie di spazi
		if (*m == ':') {
			m++;
		}
		while (*m == ' ') {
			m++;
		}
		sscanf (m, "%d,",&linnum);
		while ((*m >= '0' && *m <= '9') || *m == ',' | *m ==  ' ') {
			m++;
		}
		while ((*(--t) == ' ' || *t == ',') && t > s) {
			*t = 0;
		}
		//Fill the TreeView's model
		Gtk::TreeModel::Row row = *(errList->append());
		row[m_Columns.m_col_line] = linnum;
		row[m_Columns.m_col_type] = s;
		row[m_Columns.m_col_error] = m;
	}
}

void ErrorDialog::doSelect (bool raise)
{
	Gtk::TreeModel::iterator iter = refTreeSelection ->get_selected();
	if(iter) { //If anything is selected
		Gtk::TreeModel::Row row = *iter;
		if (row[m_Columns.m_col_line] > 0) {
			parent -> lineSelect (row[m_Columns.m_col_line], raise);
		}
	}
}

void ErrorDialog::select_child ()
{
	doSelect (false);
} 

void ErrorDialog::button_press_event (GdkEventButton *b)
{
	if (b && b -> type == GDK_2BUTTON_PRESS) {
		doSelect (true);
	}
}

void ErrorDialog::display ()
{
	show ();
	raise ();
}

bool ErrorDialog::hasErrors ()
{
Gtk::TreeModel::Children children = errList->children();
	return children.begin() != children.end();
}

ErrorDialog::ErrorDialog (EditWindow *p) :
  vbox (false, 5),
  action (false, 5),
  CloseButton ("Chiudi"),
  GoTo ("Vai all'errore")
{
	parent=p;

	set_title ("Errori");
    add (vbox);

	refTreeSelection = view.get_selection();
	refTreeSelection -> signal_changed().connect(sigc::mem_fun(*this,&ErrorDialog::select_child));
    view.set_events (Gdk::BUTTON_PRESS_MASK);
	view.signal_button_press_event().connect_notify (sigc::mem_fun(*this,&ErrorDialog::button_press_event));

	set_default_size (450,150);
	vbox.pack_start(scroller, true, true,0);
    vbox.pack_start (action, false, true, 0);

	//Add the TreeView, inside a ScrolledWindow, with the button underneath:
	scroller.add(view);

	//Only show the scrollbars when they are necessary:
	scroller.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
	
	//Create the Tree model:
	errList = Gtk::ListStore::create(m_Columns);
	view.set_model(errList);
  
	//Add the TreeView's view columns:
	view.append_column("Linea", m_Columns.m_col_line);
	view.append_column("Tipo", m_Columns.m_col_type);
	view.append_column("Errore/Warning", m_Columns.m_col_error);

	action.pack_start(CloseButton, false,false,0);
	action.pack_start(GoTo, false,false,0);

	CloseButton.signal_pressed().connect(sigc::mem_fun(*this,&ErrorDialog::CancelButton));
	GoTo.signal_pressed().connect(sigc::mem_fun(*this,&ErrorDialog::GoToButton));

	show_all ();
}
