/***************************************************************************
 *  i8052.cc
 *
 *  Sun Nov 14 10:49:20 2004
 *  Copyright  2004 Ing. Stefano Salvi
 *  Email stefano@salvi.mn.it
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>

#include "dmemory.h"
#include "xmemory.h"
#include "cmemory.h"
#include "messageinterface.h"
#include "serial.h"
#include "sfrnames.h"

#include "i8052.h"

// Contiene la dichiarazione  e l'inizializzazione di "IstSet"
#include "assembler.h"

parseNode *i8052::parseTree = NULL;
symtabNode *i8052::genSymTab = NULL;

#define BUFLEN	512

#define HEX(d) (((d) >= '0' && (d) <= '9') ? (d) - '0' : (((d) >= 'a' && (d) <= 'f') ? \
  (d) - 'a' + 10 : (((d) >= 'A' && (d) <= 'F') ? (d) - 'A' + 10 : 0)))

void i8052::setStatus (stat_val s)
{
	status = s;
	dmemory->setStatus (s);
	xmemory->setStatus (s);
	cmemory->setStatus (s);
	SFRmemory->setStatus (s);
}

void i8052::redrawIfChanged ()
{
	dmemory->redrawIfChanged ();
	xmemory->redrawIfChanged ();
	cmemory->redrawIfChanged ();
	SFRmemory->redrawIfChanged ();
}

void *i8052::Emulate_wrapper (void *th)
{
    return ((i8052 *)th) -> emulate ();
}

void i8052::run (int where)
{
char gobuf [6];
int i;
	if (Serial_Container::s->isConnected () && status != running) {
		SFRmemory->store (where);
		sprintf (gobuf,"G%04X",where);
		Serial_Container::s->schPuts (gobuf);	// Comando 'go'
		// Risucchia l'eco dei cinque caratteri stampati
		for (i = 0; i<5; i++) {
			if (Serial_Container::s->schGetc () < 0) {
        		break;
			}
		}
		Serial_Container::s->schRelease ();
		interface -> setRunningMessage  ("In Esecuzione");
		setStatus (running);
	} else {
		SFRmemory -> SetPc (where);
		setStatus (emulating);
		int err = pthread_create (&Emulate_thread, NULL, Emulate_wrapper, this);
		if (err)
		{
			printf ("Non posso creare il thread del terminale (%d) \n",err);
			exit (1);
		}
		interface -> startEmulation ();
	}
}

void i8052::setPPI (ppiinterface *c_PPI)
{
	SFRmemory-> setPPI(c_PPI);
	xmemory-> setPPI(c_PPI);
}

/*
 *	Tentativo di gestione dei tempi:
 * - Ad ogni istruzione si associa il suo numero di cicli
 * -	Si crea un contatore globale di cicli, ai quali si aggungono
 *		i cicli di ogni singola istruzione.
 * -	Ipotizziamo una frequenza di 12 MHz, quindi 1 ciclo al microsencondo
 * -	Ogni 50 mS (50000 cicli) confrontiamo il tempo emulato con il reale
 *		e se l'emulatore anticipa, usando nanosleep, lo rimettiamo in passo
 */
#define	SYNC_CYCLES	50000
#define	SYNC_MILLIS	(SYNC_CYCLES/1000)
void * i8052::emulate (void)
{
struct timeval tv;
struct timezone tz;
time_t start_second;
int milliseconds;
int totdelay;
unsigned cycles;

	if ((Serial_Container::s->isConnected () && status != running) || status != emulating) {
		return NULL;
	}

 	gettimeofday(&tv, &tz);
	start_second = tv.tv_sec;
	milliseconds = tv.tv_usec / 1000;
	totdelay = 0;
	cycles = 0;

	while (status == emulating)
	{
		if (cycles > SYNC_CYCLES) {
		int curmillis;
			cycles -= SYNC_CYCLES;
			milliseconds += SYNC_MILLIS;
 			gettimeofday(&tv, &tz);
			curmillis = (tv.tv_sec - start_second) * 1000 + tv.tv_usec / 1000;
			if (milliseconds > curmillis) {
			struct timespec t;
				totdelay += milliseconds - curmillis;
				// Deve introdurre un ritardo
				t.tv_sec = 0;
				t.tv_nsec = (milliseconds - curmillis) * 1000000;
				nanosleep (&t, NULL);
			}
		}
		cycles += realStep ();
		if (SFRmemory -> isBrk ()) {
			stop ();
			break;
		}
		sched_yield ();
	}
	printf ("Tempo programma %d millisecondi\nRitardi per correzione %d millisecondi (%2.2f%%)\n",
		milliseconds, totdelay, ((float)(milliseconds - totdelay))/milliseconds * 100 );
	return NULL;
}

int i8052::realStep ()
{
unsigned PC;
int time = IstSet [cmemory -> GetDato (PC)].getCycles ();
	// Simula il flag di trasmissione seriale
	if (SFRmemory -> testSerial ()) {
		dmemory -> SetBit (bTI,1);
	}
	PC = SFRmemory -> GetPc ();
	PC = IstSet [cmemory -> GetDato (PC)].Exec (PC) & 0xffff;
	SFRmemory -> SetPc (PC);
	return time;
}

void i8052::step ()
{
char result [80];
	if (Serial_Container::s->isConnected () && status != running)
	{
		SFRmemory->store (0);
		Serial_Container::s->schPuts ("S");	// Comando 'STEP'
		Serial_Container::s->schGets (result,79);
		fprintf (stderr, "%s\n", result);
		Serial_Container::s->schPrompt ();
	} else if (status != running) {
		realStep ();
	}
	dmemory -> redraw ();
	xmemory -> redraw ();
	cmemory -> show  (SFRmemory -> GetPc());
}

void i8052::doUpload (char *file)
{
int hexfile;
char buf [BUFLEN];
int letti,i;
int address = 0, data = 0, count = 0;
enum { WCOLON, WCNTH, WCNTL, WADH, WADM0, WADM1, WADL, WDUMH, WDUML, WDATH, WDATL, END} upload_status;

	/* Se e' in esecuzione, non deve fare l'upload !! */
	if (status == running) {
    	return;
	}
    
	hexfile = open (file,O_RDONLY);
	if (hexfile == EOF)
	{
		MessageInterface::doYesDialog ("Upload...", NULL, "Non trovo il file '%s' da scaricare", file);
		return;
	}
	
	Serial_Container::s->schPuts ("L\r");	// Comando 'upload'
	upload_status = WCOLON;
	do {
		letti = read (hexfile,buf,BUFLEN);
		for (i = 0;i < letti;i++)
		{
			if (buf [i] != '\r')
			{
				if (buf [i] == '\n') {
					Serial_Container::s->schPuts ("\r");
				}
				Serial_Container::s -> schSendc (buf [i]);
			}
			switch (upload_status)
			{
				case WCOLON:
					if (buf [i] == ':') {
						upload_status = WCNTH;
					}
					break;
				case WCNTH:
					count = HEX (buf [i]) << 4;
					upload_status = WCNTL;
					break;
				case WCNTL:
					count |= HEX (buf [i]);
					if (count) {
						upload_status = WADH;
					} else {
						upload_status = END;
					}
					break;
				case WADH:
					address = HEX (buf [i]) << 12;
					upload_status = WADM0;
					break;
				case WADM0:
					address |= HEX (buf [i]) << 8;
					upload_status = WADM1;
					break;
				case WADM1:
					address |= HEX (buf [i]) << 4;
					upload_status = WADL;
					break;
				case WADL:
					address |= HEX (buf [i]);
					upload_status = WDUMH;
					break;
				case WDUMH:
					upload_status = WDUML;
					break;
				case WDUML:
					upload_status = WDATH;
					break;
				case WDATH:
					data = HEX (buf [i]) << 4;
					upload_status = WDATL;
					break;
				case WDATL:
					data |= HEX (buf [i]);
					cmemory -> SetDato (address++, data);
					if (--count) {
						upload_status = WDATH;
					} else {
						upload_status = WCOLON;
					}
					break;
				default:
					break;
			}
		}
	} while (letti==BUFLEN);
	if (Serial_Container::s->schWaitc ('\r') ||	// scarta il comando
    	Serial_Container::s->schGets (buf,79) || Serial_Container::s->schPrompt ())
  	{
		MessageInterface::doYesDialog ("Scaricamento File" , NULL, "La scheda non risponde");
	} else {
		if (buf [0] == '?') {
			MessageInterface::doYesDialog ("Scaricamento File" , NULL, "Errore nel file '%s'" , buf);
		} else {
			interface -> setMessage (buf);
		}
	}
	xmemory -> redraw ();
	cmemory -> redraw ();
}

unsigned i8052::GetMnemonic (int pc, char *mnemonic)
{ 
	return  IstSet [cmemory->GetDato (pc)].GetMnemonic (pc, mnemonic);
}

void i8052::printError (bool warning, int linnum, errorInterface *msg, hexListFile *f)
{
char errmsg [80];
	sprintf (errmsg, "%s, line %d, %s", (warning) ? "Warning" : "Errore" , linnum, OpCode::ErrStr (OpCode::GetErr ()));
	if (f) {	// Stampa il messaggio prima del dialog, che lo "consuma"
		f -> error (errmsg);
	}
	msg->AddError(errmsg);
}

char *i8052::skipBlanks (char *s)
{
	while (*s == ' ' || *s == '\t') {
		s ++;
	}
	return s;
}

char *i8052::skipSymbol (char *s)
{
	if (issym(*s)) {
		s++;
		while (issym_cont(*s)) {
			s ++;
		}
	}
	return s;
}

#define LINELEN	128
bool i8052::pass (const char *source, bool second, errorInterface *msg, hexListFile *f)
{
FILE *assembler = fopen (source,"r");
char line [LINELEN];
int linnum = 1;
bool errors = false;
bool bitAddr;
unsigned Pc = 0;
	if (assembler) {	// Parse del file
		if (f) {
			if (!f -> open (source)) {
				msg->AddError("Error, line 0, Non posso creare il file listing o il file assemblato");
			}
		}
		for (;!feof (assembler) && !ferror (assembler); linnum ++) {
		char *s, *e;
		int j;
			OpCode::SetErr (OpCode::noerr);	// Azzera l'errore prima di iniziare
			if (f) {	// Se c'e' il file, scarica la linea precedente.
				f -> newLine(line);
			}
			fgets (line, LINELEN-1, assembler);
			s = skipBlanks (line);
			if (*s > ' ' && *s != ';') {	// Se non e' un commento o vuota
				if (issym(*s)) {	// Inizio corretto di simbolo
				unsigned char code [3];
				opcContent *ists;
					
					// Ricerca eventuale label
					e = skipSymbol (s);
					if (*e == ':') {	// Simbolo da aggungere alla symbol table
						if (!second) {	// Aggiunge simbolo a symbol table
							if (!addSymbol (s, Pc)) {
								OpCode::SetErr (OpCode::dupsym);	// Simbolo duplicato
								printError (false, linnum, msg, f);
								errors = true;
							}
						}
						s = skipBlanks (e + 1);
						e = skipSymbol (s);
						if (*s < ' ' || *s == ';') {	// Fine riga
							continue;
						}
					}
				
					j = commonAsm (s, Pc, code, ists, second);
					if (ists) {
						if (ists -> getType () == opcode) {
							// Tutto OK, nessun errore
							if (second && !OpCode::GetErr ()) {	// Seconda passata e niente errori
							int i;
								for (i = 0; i < j; i++) {
									f -> add(Pc + i, code [i]);
								}
							}
							Pc += j;
						} else {
							s = skipBlanks (e);
							OpCode::ResetErr ();	// Elimina l'errore "opcode non riconosciuto" per pseudo
							switch (ists -> getType ()) {
								case org :
									Pc = expr(s, bitAddr, Pc, true);	// Deve essere definito anche nella prima passata
									if (bitAddr) {
										OpCode::SetErr (OpCode::isbitaddr);
									}
									break;
								case equ :
									if (!second) {	// Aggiunge simbolo a symbol table
										e = skipSymbol (s);
										e = skipBlanks (e);
										if (*e != ',') {
											OpCode::SetErr (OpCode::noop);	// Manca il parametro
										} else {
											e = skipBlanks (e+1);
											j = expr (e, bitAddr, Pc, second);
											if (!addSymbol (s, j)) {
												OpCode::SetErr (OpCode::dupsym);	// Simbolo duplicato
											}
										}
									}
									break;
								case db :	// lista di stringhe circondate da " oppure di valori separati da ,
									while (OpCode::GetErr () == OpCode::noerr) {
										e = skipBlanks (e);
										if (*e == '"') {
											e++;	// salta il "
											while ((*e >= ' ' || *e == '\t') && *e != '"') {
												if (*e == '\\') {
													switch ( *++e) {
														case 'n' :
															j = '\n';
															break;
														case 'r' :
															j = '\r';
															break;
														case 't' :
															j = '\t';
															break;
														case 'f' :
															j = '\f';
															break;
														default:
															j = *e;
													}
													if (f) {
														f -> add(Pc, j & 0xff);
													}
													Pc++;
													e++;
												} else {
													if (f) {
														f -> add(Pc, *e & 0xff);
													}
													e++;
													Pc++;
												}
											}
											if (*e == '"') {	// Fine stringa: c'e' il "?
												e++;
											} else {	// Errore
												OpCode::SetErr (OpCode::strerr);
											}
										} else {
											j = expr (e, bitAddr, Pc, second);
											if (OpCode::GetErr () >= OpCode::nonfatal) {
												break;
											}
											if ((j	> 255 || j < -128) && second) {
												OpCode::SetErr (OpCode::toobig);
											}
											if (f) {
												f -> add(Pc, j & 0xff);
											}
											Pc++;
										}
										e = skipBlanks (e);
										if (*e == ',') {
											e ++;
										} else {
											break;
										}
									}
									break;
								case dw :	// Lista di valori separati da virgola
									while (OpCode::GetErr () == OpCode::noerr) {
										e = skipBlanks (e);
										j = expr (e, bitAddr, Pc, second);
										e = skipBlanks (e);
										if (OpCode::GetErr () >= OpCode::nonfatal) {
											break;
										}
										if (f) {
											f -> add(Pc, (j >> 8) & 0xff);
											f -> add(Pc + 1, j & 0xff);
										}
										Pc += 2;
										if (*e == ',') {
											e ++;
										} else {
											break;
										}
									}
									break;
								case df :
									break;
								case skip :
									Pc += expr(s, bitAddr, Pc, true);	// Deve essere definito anche nella prima passata
									break;
								case end :
								default :
									break;
							}
						}
					}
					if (OpCode::GetErr ()) {
						printError (false, linnum, msg, f);
						errors = true;
					}
				} else {
					OpCode::SetErr (OpCode::noopcode);
					printError (true, linnum, msg, f);
				}
			}
		}
		fclose (assembler);
		return !errors;
	} else {
		msg->AddError("Warning, line 0, Manca il file da assemblare");
	}
	return false;
}

void i8052::assembler (const char *source, errorInterface *msg)
{
	if (pass (source, false, msg, NULL)) {
	hexListFile f;
		pass (source, true, msg, &f);
	}
	clearSymTab ();
}

bool i8052::cmpSym (char *&s,const char *d)
{
char *e = s;
	while (toupper (*e) == *d) {
		e++;
		d++;
	}
	if (*d || issym_cont(*e)) {	// Non e' terminato il simbolo oppure il confronto
		return false;
	}
	s = e;
	return true;
}

int i8052::commonAsm (char *buf,unsigned Pc, unsigned char *code, opcContent *&ists, bool secondpass)
{
int j = 0, i, n;
unsigned char ad;
OpCode::argStore pars [3];	// Array dei parametri dell'istruzione
OpCode::ErrVals e;
bool bitAddr;
	if (!*buf) {
		return 0;
	}
	OpCode::ResetErr ();
	ists = parseTree->find (buf);
	if (ists) {	// se ha trovato l'opcode
		if ((n = ists->getNum ())) {
			buf =  skipSymbol (buf);	// Si posiziona dopo l'opcode
			/* Trova i parametri dell'istruzione e li mette nell'array */
			for (i = 0; i < 3; i++) {	// Inizializza l'array
				pars [i].type = OpCode::noParg;
			}

			for (i = 0; true ; i++) {
			char *t;
				buf = skipBlanks (buf);
				if (!i && (*buf == ';' || *buf < ' ' )) {	// Nessun parametro
					break;
				}
				switch (*buf) {
					case '#':	// Immediato
						pars [i].type = OpCode::Pimm;
						buf = skipBlanks (buf+1);	// Salta la '#'
						pars [i].val = expr (buf, bitAddr, Pc, secondpass);
						break;
					case '/':	// Bit invertito
						pars [i].type = OpCode::notBitDir;
						buf = skipBlanks (buf+1);	// Salta la '/'
						t = OpCode::Syms [1].GetCode (buf,&ad);
						if (t) {	// e' un simbolo predefinito - usa
							buf = t;
							pars [i].val = ad;
						} else {
							pars [i].val = expr (buf, bitAddr, Pc, secondpass);
						}
						break;
					case '@':	// indReg, indDptr, idxPc, idxDptr
						switch (*++buf) {
							case 'a':
							case 'A':	// A+DPTR / A+PC
								buf = skipBlanks (buf+1);	// Salta la 'A'
								if (*buf != '+') {
									OpCode::SetErr (OpCode::inderr);
									break;
								}
								buf = skipBlanks (buf+1);	// Salta il '+'
								if (cmpSym (buf,"PC")) {
									pars [i].type = OpCode::idxPc;
								} else if (cmpSym (buf,"DPTR")) {
									pars [i].type = OpCode::idxDptr;
								} else {
									OpCode::SetErr (OpCode::idxerr);
								}
								break;
							case 'r':
							case 'R':	// R0 / R1
								if (issym_cont(buf[2])) {	// Se il simbolo e' maggiore di 2 caratteri (Rx)
									OpCode::SetErr (OpCode::idxerr);
								} else {
									buf ++;	// Salta la "R"
									if (*buf == '0' || *buf == '1') {	// Corretto
										pars [i].type = OpCode::indReg;
										pars [i].val = *buf++ - '0';
										break;
									} else if (*buf >= '2' && *buf <= '7') {
										OpCode::SetErr (OpCode::indreg);
									}
									OpCode::SetErr (OpCode::idxerr);
								}
								break;
							case 'd':
							case 'D':	// DPTR
								if (cmpSym (buf,"DPTR")) {
									pars [i].type = OpCode::indDptr;
								} else {
									OpCode::SetErr (OpCode::idxerr);
								}
								break;
							default:
								OpCode::SetErr (OpCode::inderr);
								break;
						}
						break;
					case 'r':
					case 'R':
					case 'a':
					case 'A':
					case 'c':
					case 'C':
					case 'd':
					case 'D':
						if (*buf == 'r' || *buf == 'R') {
							if (!issym_cont(buf[2])) {	// e' di due lettere, potrebbe essere un registro
								if (buf [1] >= '0' && buf[1] <= '7') {	// Corretto
									pars [i].type = OpCode::reg;
									buf++;	// Salta la 'R'
									pars [i].val = *buf++ - '0';
									break;
								}
							}
						} else if (*buf == 'a' || *buf == 'A') {
							if (!issym_cont(buf[1])) {	// e' di una lettera, quindi e' 'A'
									pars [i].type = OpCode::aReg;
									buf++;	// Salta la 'A'
									break;
							} else if (!issym_cont(buf[2]) && toupper (buf[1]) == 'B') {	// e' di due lettere, e la seconda e' 'B' -> 'AB'
									pars [i].type = OpCode::bReg;
									buf+=2;	// Salta la 'A' e la 'B'
									break;
							}
						} else if (*buf == 'c' || *buf == 'C') {
							if (!issym_cont(buf[1])) {	// e' di una lettera, quindi e' 'C'
									pars [i].type = OpCode::cReg;
									buf++;	// Salta la 'C'
									break;
							}
						} else {	// qui non puo' che essere 'D'
							if (cmpSym (buf,"DPTR")) {
								pars [i].type = OpCode::dptrReg;
								break;
							}
						}
					default:
						// Controlla se e' un SFR byte.
						t = OpCode::Syms [0].GetCode (buf,&ad);
						if (t) {	// e' un simbolo predefinito - usa
							buf = t;
							pars [i].type = OpCode::dir;
							pars [i].val = ad;
						} else {
							t = OpCode::Syms [1].GetCode (buf,&ad);
							if (t) {	// e' un simbolo predefinito - usa
								buf = t;
								pars [i].type = OpCode::bitDir;
								pars [i].val = ad;
							} else {
								pars [i].val = expr (buf, bitAddr, Pc, secondpass);
								pars [i].type = (bitAddr) ? OpCode::bitDir : OpCode::dir;
							}
						}
						break;
				}
				buf = skipBlanks (buf);
				if (OpCode::GetErr () >= OpCode::nonfatal) {
					break;
				}
				if (*buf == ',') {
					buf ++;
					if (i >= 2) {	// Troppi parametri
						OpCode::SetErr (OpCode::parserr);
						break;
					}
				} else {
					break;
				}
			}

			for (i = 0; i < n; i++) {
				j = (*ists)[i];
		    	j = IstSet [j].Asm (j,pars,Pc, code, secondpass);
				e = OpCode::GetErr ();
				if (j) {
		      		break;
				}
			}
			if (i >= n) {	// Nessuna versione corrisponde
				OpCode::SetErr (OpCode::parserr);
				return 0;
			}
		}
	} else {
		OpCode::SetErr (OpCode::noopcode);
	}
	return j;
}

/* Ritorna la lunghezza in byte dell'istruzione
 * assemblata, o 0 se errata */
int i8052::Asm (char *buf,unsigned Pc)
{
unsigned char code [3];
opcContent *ists;
int i,j;
	j = commonAsm (buf, Pc, code, ists, true);
	if (OpCode::GetErr ()) {
		MessageInterface::doYesDialog ("Assembler: Errore" , NULL, "Errore :\n%s.",
			OpCode::ErrStr (OpCode::GetErr ()));
		return 0;
	}
	// Tutto OK, nessun errore
	for (i = 0; i < j; i++) {
		cmemory -> SetDato (Pc + i, code [i]);
		cmemory -> saveValue (Pc + i);
	}
	return j;
}

void i8052::reload ()
{
	dmemory -> reload ();
	dmemory -> SFRreload ();
	xmemory -> reload ();
	cmemory -> reload ();
}


/* (Vado a memoria - solo matematiche)
 * <val> ::= '$' | <sym> | <num> | '(' <addexpr> ')'
 * <unexpr> ::= [ '~' | '-' ] <val>
 * <mulexpr> ::= <unexpr> [ ( '*' | '/' | '&' | '%' ) <unexpr>](rep)
 * <addexpr> ::= <mulexpr> [ ( '+' | '-' | '|' ) <mulexpr>](rep)
 * <expr> ::= <addexpr>
 */

/* Estrae il valore numerico dalla stringa
 * Ritorna il carattere seguente al numero riconosciuto
 */
int i8052::GetInt (char *&a)
{
int i;
int num;
	if (isdigit (*a)) {
	    // Raggiunge la fine del numero
    	for (i=0;isxdigit (a[i]) || a[i] == 'H' || a[i] == 'h' || a[i]=='O' || a[i]=='o';i++);
		num = 0;
		switch (a [i - 1])	// a [i-1] = simbolo della base (h,d,o,b)
		{
			case 'B' :
			case 'b' :      // Binario
				for (;*a >= '0' && *a <= '1';a++)
				{
					num *= 2;
					num += *a - '0';
				}
				a++;	// Salta la "b"
				break;
			case 'H' :
			case 'h' :      // Esadecimale
				for (;isxdigit (*a);a++)
				{
					num *= 16;
					if (*a <= '9') {
						num += *a - '0';
					} else if (*a <= 'F') {
						num += *a - 'A' + 10;
					} else {
						num += *a - 'a' + 10;
					}
				}
				a++;	// Salta la "h"
				break;
			case 'O' :
			case 'o' :      // Ottale
				for (;*a >= '0' && *a <= '7';a++)
				{
					num *= 8;
					num += *a - '0';
				}
				a++;	// Salta la "o"
				break;
			default :       // Decimale
				for (;isdigit (*a);a++)
				{
					num *= 10;
					num += *a - '0';
				}
				if (*a == 'D' || *a == 'd') {
					a++;
				}
				break;
		}
	} else {
		OpCode::SetErr (OpCode::numerr);
		return 0;
	}
	return num;
}

int i8052::bitAddress (char *&s, int num, bool &bitAddr, bool second)
{
int bit;
	if (*s == '.') {	// E' un'indirizzo 'bit'
		bitAddr = true;
		s ++;
		if (*s >= '0' && *s <= '7' && !issym_cont(s [1])) {
			bit = *s++ - '0';
			if ((num >= 0x20 && num < 0x30) || (num >= 0x80 && num % 8 == 0)) {	// indirizzo byte valido
				if (num < 0x80) {	// Variabli bitadressable
					num = (num - 0x20) * 8;
				}
				num |= bit;
			} else {
				OpCode::SetErr (OpCode::notbitadressable);
			}
		} else {
			OpCode::SetErr (OpCode::bitnumerr);
			s = skipSymbol (s);
			return 0;
		}
	}
	return num;
}

int i8052::val (char *&s, bool &bitAddr, unsigned pc, bool second)
{
int num;
	if (*s == '$') {
		return pc;
	} else if (*s == '(') {
		s = skipBlanks (s + 1);
		num = addexpr (s, bitAddr, pc, second);
		if (*s == ')') {
			s = skipBlanks (s + 1);
		} else {
			OpCode::SetErr (OpCode::parnotbal);
		}
	} else if (*s == '\'') { // costante Carattere
		if (*s == '\\') {
			switch (*++s) {
				case 'n' :
					num = '\n';
					break;
				case 'r' :
					num = '\r';
					break;
				case 't' :
					num = '\t';
					break;
				case 'f' :
					num = '\f';
					break;
				default:
					num = *s;
					break;
			}
		} else {
			num = *++s;
		}
		if (*++s != '\'') {
			OpCode::SetErr (OpCode::charconst);
			return 0;
		} else {
			s++;
		}
	} else  if (isdigit (*s)) {	// Costante numerica
		num = GetInt (s);
		num = bitAddress (s, num, bitAddr, second);
	} else {	// Simbolo
		num = 0;
		if (genSymTab) {
		symContent *content = genSymTab -> find (s);
			if (content && content->isPresent ()) {
				num = content->get ();
			} else if (second) {
				OpCode::SetErr (OpCode::nosym);	// simbolo non trovato
			}
		} else if (second) {
			OpCode::SetErr (OpCode::nosym);	// simbolo non trovato (symbol table vuota)
		}
		s = skipSymbol (s);
		num = bitAddress (s, num, bitAddr, second);
	}
	s = skipBlanks (s);
	return num;
}

 /* <unexpr> ::= [ '~' | '-' ] <val> */
int i8052::unexpr (char *&s, bool &bitAddr, unsigned pc, bool second)
{
int num;
	if (*s == '~') {
		s = skipBlanks (s + 1);
		num = (~ val (s, bitAddr, pc, second) & 0xffff);	// Limita a 16 bit
		if (bitAddr) {
			OpCode::SetErr (OpCode::illbitaddrop);	// simbolo non trovato
		}
	} else if (*s == '-') {
		s = skipBlanks (s + 1);
		num = - val (s, bitAddr, pc, second);
		if (bitAddr) {
			OpCode::SetErr (OpCode::illbitaddrop);	// simbolo non trovato
		}
	} else {
		num = val (s, bitAddr, pc, second);
	}
	return num;
}

/* <mulexpr> ::= <unexpr> [ ( '*' | '/' | '&' | '%' ) <unexpr>](rep) */
int i8052::mulexpr (char *&s, bool &bitAddr, unsigned pc, bool second)
{
int num = unexpr (s, bitAddr, pc, second);
	if (!OpCode::GetErr ()) {
		while (*s == '*' || *s == '/' || *s == '&' || *s == '%') {
		int num2;
		char oper = *s;
			if (bitAddr) {
				OpCode::SetErr (OpCode::illbitaddrop);	// indirizzo di bit in espressione
				return 0;
			}
			s = skipBlanks (s + 1);
			num2 = unexpr (s, bitAddr, pc, second);
			if (bitAddr) {
				OpCode::SetErr (OpCode::illbitaddrop);	// indirizzo di bit in espressione
			}
			if (OpCode::GetErr ()) {
				return 0;
			}
			switch (oper) {
				case '*':
					num *= num2;
					break;
				case '/':
					if (num2) {
						num /= num2;
					} else if (second) {
						OpCode::SetErr (OpCode::divbyzero);	// Divisione per 0
					}
					break;
				case '&':
					num &= num2;
					break;
				case '%':
					if (num2) {
						num %= num2;
					} else if (second) {
						OpCode::SetErr (OpCode::divbyzero);	// Divisione per 0
					}
					break;
			}
		}
	}
	return num;
}

/* <addexpr> ::= <mulexpr> [ ( '+' | '-' | '|' ) <mulexpr>](rep) */
int i8052::addexpr (char *&s, bool &bitAddr, unsigned pc, bool second)
{
int num = mulexpr (s, bitAddr, pc, second);
	if (!OpCode::GetErr ()) {
		while (*s != ')' && *s != ';' && *s != ',' && *s >= ' ') {	// Se c'e' ancora continuazione di linea
		int num2;
		char oper = *s;
			if (bitAddr) {
				OpCode::SetErr (OpCode::illbitaddrop);	// indirizzo di bit in espressione
				return 0;
			}
			s = skipBlanks (s + 1);
			num2 = mulexpr (s, bitAddr, pc, second);
			if (bitAddr) {
				OpCode::SetErr (OpCode::illbitaddrop);	// indirizzo di bit in espressione
			}
			if (OpCode::GetErr ()) {
				return 0;
			}
			switch (oper) {
				case '+':
					num += num2;
					break;
				case '-':
					num -= num2;
					break;
				case '|':
					num |= num2;
					break;
				default:	// Operatore o simbolo non riconosciuto
					OpCode::SetErr (OpCode::illoper);	// Operatore illegale
			}
		}
	}
	return num;
}

int i8052::expr (char *&s, bool &bitAddr, unsigned pc, bool second)
{
int num;
	bitAddr = false;
	num = addexpr (s, bitAddr, pc, second);
	if (*s == ')') {	// Se c'e' ancora continuazione di espressione
		OpCode::SetErr (OpCode::parnotbal);	// Operatore illegale
	}
	if (num > 0xffff || num < -32767) {
		OpCode::SetErr (OpCode::toobig);
	}
	return num;
}
/* ---- Fine valutatore espressione ---- */

i8052::i8052 (Serial_Container *i)
{
	interface = i;
	// Crea le aree di memoria
	SFRmemory = new SFRMemory (i);
	dmemory = new DataMemory (0,256, SFRmemory);
	xmemory = new XDataMemory (0x8050, XDataMemory::size);
	cmemory = new CodeMemory (*xmemory);
	OpCode::setProcessor (dmemory, xmemory, cmemory); 
	setStatus (stopped);
	if (!parseTree) {
	int i;
		for (i = 0; i < 256;i ++) {
			IstSet [i].add (parseTree, i);
		}
printf ("Numero nodi %d, massimo numero di opcpde per nodo %d\n",parseNode::nnodes ,parseNode::maxopcode);
		parseNode::addToRoot (parseTree, ".ORG", org, 0);
		parseNode::addToRoot (parseTree, ".EQU", equ, 0);
		parseNode::addToRoot (parseTree, ".DB", db, 0);
		parseNode::addToRoot (parseTree, ".DW", dw, 0);
		parseNode::addToRoot (parseTree, ".DS", skip, 0);
		parseNode::addToRoot (parseTree, ".BYTE", db, 0);
		parseNode::addToRoot (parseTree, ".WORD", dw, 0);
		parseNode::addToRoot (parseTree, ".FLAG", df, 0);
		parseNode::addToRoot (parseTree, ".SKIP", skip, 0);
		parseNode::addToRoot (parseTree, ".END", end, 0);
	}
}

/*
 * hexListFile
 */

char *hexListFile::changeExt (const char *name, const char *ext)
{
char *slash, *dot, *s = new char [strlen (name) + 4];
	strcpy (s, name);
	dot = strrchr (s,  '.');
	if (dot) {
		slash = strrchr (s, '/');
		if (slash && slash > dot) {
			dot = s + strlen (s);
		}
	} else {
		dot = s + strlen (s);
	}
	strcpy (dot, ext);
	return s;
}

bool hexListFile::open (const char *name)
{
char *hexname = changeExt (name, ".hex");
char *lstname = changeExt (name, ".lst");
bool result = true;
printf ("Hex '%s'\nLst '%s'\n", hexname, lstname);
	
	lst = fopen (lstname, "w");
	if (lst) {
		hex = fopen (hexname, "w");
		if (!hex) {
			fclose (lst);
			lst = NULL;
			result = false;
		}
	} else {
		result = false;
	}

	delete hexname;
	delete lstname;
	return result;
}

void hexListFile::add (unsigned ad, unsigned char val)
{
	if (curad != ad) {	// Cambio origine
		flush ();
	}
	buf [hexPending++] = val;
	curad = ad + 1;
	if (hexPending >= 16) {
		flush ();
	}
	if (!lstPending++) {
		fprintf (lst,"%04X ", ad);
	}
	fprintf (lst,"%02X ", val);
	if (lstPending >= 4) {
		newLine (NULL);
	}
}

void hexListFile::newLine (char *l)
{
int i;
	if (!lstPending) {
		if (!src) {	// Ne linea ne istruzioni, salta
			src = l;
			return;
		}
		fputs ("     ", lst);
	}
	for (i = lstPending; i < 4; i++)
	{
		fputs ("   ", lst);
	}
	fputs ( (src) ? src : "\n", lst);
	src = l;
	lstPending = 0;
}

void hexListFile::error (char *e)
{
	newLine (NULL);
	fprintf (lst, "**** %s ****\n", e);
}

/* Formato hex:
 * :<numchars><addr hi><addr lo>00<char>..<char><chks>
 */
void hexListFile::flush ()
{
unsigned i,chks = hexPending;
	if (hexPending) {
	int ad= curad - hexPending;
		chks += ad & 0xff;
		chks += ad >> 8;
		fprintf (hex,":%02X%04X00", hexPending, ad);
		for (i = 0; i < hexPending; i++)
		{
			fprintf (hex,"%02X", buf [i]);
			chks += buf [i];
		}
		fprintf (hex,"%02X\n", (- chks) & 0xff);
		hexPending = 0;
	}
}

hexListFile::~hexListFile ()
{
	if (hexPending) {
		flush ();
	}
	fputs (":00000001FF\n", hex);
	if (lstPending) {
		newLine (NULL);
	}
	if (lst) {
		fclose (lst);
	}
	if (hex) {
		fclose (hex);
	}
}
