/***************************************************************************
 *  memorypanel.cc
 *
 *  Sun Nov 14 10:49:20 2004
 *  Copyright  2004 Ing. Stefano Salvi
 *  Email stefano@salvi.mn.it
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 * Deve avere "native scrolling support".
 * Da "scrolledwindow.cc":
 * Per verificare se il widget inserit ha "native scrolling window" verifica se
 * esiste set_scroll_adjustments_signal; il test e'
 * "if (GTK_WIDGET_GET_CLASS(gwidget)->set_scroll_adjustments_signal == 0)"
 *
 * Veniamo ora a creare il segnale "set_scroll_adjustments_signal".
 * Per prima cosa occorre creare una "classe gtk++" per il nostro "widget".
 *
 * - Esiste, nella classe base _GtkWidgetClass un campo 
 *   'guint set_scroll_adjustments_signal;' che contiene l'ID del segnale
 * - Dato che la struttura "class" è comune a tutte le istanze dell'oggetto,
 *   occorre creare una nuova classe di oggetti per impostare il 
 *   campo set_scroll_adjustments_signal che viene testato da "ScroledWindow".
 * - Seguo quindi le istruzioni del tutorial all'indirizzo
 *   http://www.gtk.org/tutorial/sec-creatingacompositewidget.html
 * - Occorre poi creare il wrapper gtkmm
 */
 

#include <glib/gmacros.h>
#include <glib.h>
#include <gtkmm.h>
#include <libgnomemm.h>

#include <gtk/gtkobject.h>

#include "memory.h"
#include "pcmon.h"
#include "memorypanel.h"

/* ******************************************************************** *
 * Implementazione dell'oggetto GTK+ sottostante					    *
 * ******************************************************************** */
#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

static void memorypanel_class_init          (MemorypanelClass *klass);
static void memorypanel_init                (Memorypanel      *mp);
static void memorypanel_set_adjustments    (Memorypanel      *mp,
                                           GtkAdjustment  *hadj,
                                           GtkAdjustment  *vadj);
static void memorypanel_adjustment_changed (GtkAdjustment *adjustment,
			       Memorypanel     *mp);
static GtkAdjustment *new_default_adjustment (void);
static void memorypanel_set_adjustment_upper (GtkAdjustment *adj,
				 gdouble        upper,
				 gboolean       always_emit_changed);

/* in 	gtkmarshalers.c (gtkmarshalers.h) - non si riesce a trovare l'header */
void _gtk_marshal_VOID__OBJECT_OBJECT  	(   	GClosure *   	 closure,
		GValue *  	return_value,
		guint  	n_param_values,
		const GValue *  	param_values,
		gpointer  	invocation_hint,
		gpointer  	marshal_data
	);

guint
memorypanel_get_type ()
{
  static guint mp_type = 0;

  if (!mp_type)
    {
      static const GTypeInfo mp_info =
      {
		sizeof (MemorypanelClass),
		NULL,		/* base_init */
		NULL,		/* base_finalize */
		(GClassInitFunc) memorypanel_class_init,
		NULL,		/* class_finalize */
		NULL,		/* class_data */
		sizeof (Memorypanel),
		0,		/* n_preallocs */
		(GInstanceInitFunc) memorypanel_init,
      };

      mp_type = g_type_register_static (GTK_TYPE_DRAWING_AREA, "Memorypanel",
					    &mp_info, (GTypeFlags) 0);
    }

  return mp_type;
}

static void
memorypanel_class_init (MemorypanelClass *gtk_class)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;
	
  object_class = (GtkObjectClass*) gtk_class;
  widget_class = (GtkWidgetClass*) gtk_class;
  
  gtk_class->set_scroll_adjustments = memorypanel_set_adjustments;

  widget_class->set_scroll_adjustments_signal =
    g_signal_new ("set_scroll_adjustments",
		  G_OBJECT_CLASS_TYPE (object_class),
		  (GSignalFlags) (G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION),
		  G_STRUCT_OFFSET (MemorypanelClass, set_scroll_adjustments),
		  NULL, NULL,
		  _gtk_marshal_VOID__OBJECT_OBJECT,
		  G_TYPE_NONE, 2,
		  GTK_TYPE_ADJUSTMENT,
		  GTK_TYPE_ADJUSTMENT);


  gtk_class->memorypanel = NULL;
}

static GtkAdjustment *
new_default_adjustment (void)
{
  return GTK_ADJUSTMENT (gtk_adjustment_new (0.0, 0.0, 0.0, 0.0, 0.0, 0.0));
}

static void           
memorypanel_set_adjustments (Memorypanel *mp,
			    GtkAdjustment *hadj,
			    GtkAdjustment *vadj)
{
  gboolean need_adjust = FALSE;

  g_return_if_fail (IS_MEMORYPANEL (mp));

  if (hadj)
    g_return_if_fail (GTK_IS_ADJUSTMENT (hadj));
  else if (mp->hadjustment)
    hadj = new_default_adjustment ();
  if (vadj)
    g_return_if_fail (GTK_IS_ADJUSTMENT (vadj));
  else if (mp->vadjustment)
    vadj = new_default_adjustment ();
  
  if (mp->hadjustment && (mp->hadjustment != hadj))
    {
      g_signal_handlers_disconnect_by_func (mp->hadjustment,
					    (void *)memorypanel_adjustment_changed,
					    mp);
      g_object_unref (mp->hadjustment);
    }
  
  if (mp->vadjustment && (mp->vadjustment != vadj))
    {
      g_signal_handlers_disconnect_by_func (mp->vadjustment,
					    (void *) memorypanel_adjustment_changed,
					    mp);
      g_object_unref (mp->vadjustment);
    }
  
  if (mp->hadjustment != hadj)
    {
      mp->hadjustment = hadj;
      g_object_ref (mp->hadjustment);
      gtk_object_sink (GTK_OBJECT (mp->hadjustment));
      memorypanel_set_adjustment_upper (mp->hadjustment, mp->width, FALSE);
      
      g_signal_connect (mp->hadjustment, "value_changed",
			G_CALLBACK (memorypanel_adjustment_changed),
			mp);
      need_adjust = TRUE;
    }
  
  if (mp->vadjustment != vadj)
    {
      mp->vadjustment = vadj;
      g_object_ref (mp->vadjustment);
      gtk_object_sink (GTK_OBJECT (mp->vadjustment));
      memorypanel_set_adjustment_upper (mp->vadjustment, mp->height, FALSE);
      
      g_signal_connect (mp->vadjustment, "value_changed",
			G_CALLBACK (memorypanel_adjustment_changed),
			mp);
      need_adjust = TRUE;
    }

  /* vadj or hadj can be NULL while constructing; don't emit a signal
     then */
  if (need_adjust && vadj && hadj)
    memorypanel_adjustment_changed (NULL, mp);
}

/* Callbacks */

static void
memorypanel_adjustment_changed (GtkAdjustment *adjustment,
			       Memorypanel     *mp)
{
  GtkWidget *widget;

  widget = GTK_WIDGET (mp);

/*  if (mp->freeze_count)
    return; */

  if (GTK_WIDGET_REALIZED (mp))
    {
/*      gdk_window_move (layout->bin_window,
		       - layout->hadjustment->value,
		       - layout->vadjustment->value);
      
      gdk_window_process_updates (layout->bin_window, TRUE); */
    }
}

static void
memorypanel_set_adjustment_upper (GtkAdjustment *adj,
				 gdouble        upper,
				 gboolean       always_emit_changed)
{
  gboolean changed = FALSE;
  gboolean value_changed = FALSE;
  
  gdouble min = MAX (0., upper - adj->page_size);

  if (upper != adj->upper)
    {
      adj->upper = upper;
      changed = TRUE;
    }
      
  if (adj->value > min)
    {
      adj->value = min;
      value_changed = TRUE;
    }
  
  if (changed || always_emit_changed)
    gtk_adjustment_changed (adj);
  if (value_changed)
    gtk_adjustment_value_changed (adj);
}


static void
memorypanel_init (Memorypanel *mp)
{
}

#ifdef __cplusplus
}
#endif /* __cplusplus */

/* ******************************************************************** *
 * Fine Implementazione dell'oggetto GTK+ sottostante					*
 * ******************************************************************** */
/* ******************************************************************** *
 * The *_Class implementation: 											*
 * ******************************************************************** */
// for DrawingArea_Class
#include <gtkmm/private/drawingarea_p.h>

const Glib::Class& MemoryPanel_Class::init()
{
  if(!gtype_) // create the GType if necessary
  {
    // Glib::Class has to know the class init function to clone custom types.
    class_init_func_ = &MemoryPanel_Class::class_init_function;

    // This is actually just optimized away, apparently with no harm.
    // Make sure that the parent type has been created.
    //CppClassParent::CppObjectType::get_type();

    // Create the wrapper type, with the same class/instance size as the base type.
    register_derived_type(memorypanel_get_type());

    // Add derived versions of interfaces, if the C type implements any interfaces:
  }

  return *this;
}

void MemoryPanel_Class::class_init_function(void* g_class, void* class_data)
{
  BaseClassType *const klass = static_cast<BaseClassType*>(g_class);
  CppClassParent::class_init_function(klass, class_data);

  klass->set_scroll_adjustments = &set_scroll_adjustments_callback;
}


void MemoryPanel_Class::set_scroll_adjustments_callback(Memorypanel* self, GtkAdjustment* p0, GtkAdjustment* p1)
{
  CppObjectType *const obj = dynamic_cast<CppObjectType*>(
      Glib::ObjectBase::_get_current_wrapper((GObject*)self));

  // Non-gtkmmproc-generated custom classes implicitly call the default
  // Glib::ObjectBase constructor, which sets is_derived_. But gtkmmproc-
  // generated classes can use this optimisation, which avoids the unnecessary
  // parameter conversions if there is no possibility of the virtual function
  // being overridden:
  if(obj && obj->is_derived_())
  {
    try // Trap C++ exceptions which would normally be lost because this is a C callback.
    {
      // Call the virtual member method, which derived classes might override.
      obj->on_set_scroll_adjustments(Glib::wrap(p0), Glib::wrap(p1));
    }
    catch(...)
    {
      Glib::exception_handlers_invoke();
    }
  }
  else
  {
    BaseClassType *const base = static_cast<BaseClassType*>(
        g_type_class_peek_parent(G_OBJECT_GET_CLASS(self)) // Get the parent class of the object class (The original underlying C class).
    );

    // Call the original underlying C function:
    if(base && base->set_scroll_adjustments)
      (*base->set_scroll_adjustments)(self, p0, p1);
  }
}


Glib::ObjectBase* MemoryPanel_Class::wrap_new(GObject* o)
{
  return manage(new MemoryPanel((Memorypanel*)(o)));
}
/* ******************************************************************** *
 * End *_Class implementation: 											*
 * ******************************************************************** */
/* ******************************************************************** *
 * implementation (wrapper): 											*
 * ******************************************************************** */
MemoryPanel::MemoryPanel(Memorypanel* castitem) :
  Gtk::DrawingArea((GtkDrawingArea*)(castitem))
{
}

MemoryPanel::~MemoryPanel()
{
  destroy_();
}

MemoryPanel::CppClassType MemoryPanel::memorypanel_class_; // initialize static member

GType MemoryPanel::get_type()
{
  return memorypanel_class_.init().get_type();
}

GType MemoryPanel::get_base_type()
{
  return memorypanel_get_type();
}

MemoryPanel* wrap(Memorypanel* object, bool take_copy)
{
  return dynamic_cast<MemoryPanel *> (Glib::wrap_auto ((GObject*)(object), take_copy));
}

/* ******************************************************************** *
 * End implementation (wrapper): 										*
 * ******************************************************************** */

MemoryPanel::MemoryPanel (Memory *d, Pcmon *p) :
  Gtk::DrawingArea(Glib::ConstructParams(memorypanel_class_.init(), (char*) 0))
{
    parent = p;
    data = d;
    curW = 0;
    curH = 0;
    xOffs = 0;
    yOffs = 0;
    curOn = -1;
	
	HAdj = 0;
	VAdj = 0;

	d -> setRefresh (this);

	set_flags (Gtk::CAN_FOCUS);
    set_events (Gdk::EXPOSURE_MASK | Gdk::KEY_PRESS_MASK | Gdk::BUTTON_PRESS_MASK | Gdk::FOCUS_CHANGE_MASK);
    set_sensitive (true);
	
	font = new Pango::FontDescription ("Courier,monospace, 10");
	Glib::RefPtr<Pango::Layout> pangolayout = create_pango_layout("M");
	pangolayout->set_font_description (*font);
	pangolayout->get_pixel_size (text_w, text_h);

	g_object_set_data ((GObject *) gobj (), "class", this);	

    curX = curY = Zone = low = 0;
}

void MemoryPanel::on_set_scroll_adjustments(Gtk::Adjustment* hadj, Gtk::Adjustment* vadj)
{
	HAdj = hadj; // Glib::wrap (hadj);
	VAdj = vadj; // Glib::wrap (vadj);
    VAdj->signal_value_changed().connect(sigc::mem_fun(*this,&MemoryPanel::yChanged));
    HAdj->signal_value_changed().connect(sigc::mem_fun(*this,&MemoryPanel::xChanged));
    curW = get_width ();
    curH = get_height ();
	HAdj->set_lower (0.0);
	HAdj->set_step_increment (1.0);
	VAdj->set_lower (0.0);
	VAdj->set_step_increment (1.0);
    setScrollSizes (curW, curH);
}

bool MemoryPanel::on_configure_event (GdkEventConfigure *event)
{
  if (curW != (unsigned) get_width () || curH != (unsigned) get_height ())
  {
    curW = get_width ();
    curH = get_height ();
    setScrollSizes (curW, curH);
  }
  return true;
}

void MemoryPanel::setScrollSizes (int w, int h)
{
  if (VAdj) {
    VAdj->set_page_size ((gfloat)(h - 2 * border) / text_h);
    VAdj->set_page_increment ((gfloat)(h - 2 * border) / text_h / 2);
    VAdj->set_upper (data -> getSize () / 16.0);
    VAdj->changed ();
  }

  if (HAdj) {
    HAdj->set_page_size ((gfloat)(w - 2 * border) / text_w);
    HAdj->set_page_increment ((gfloat)(w - 2 * border) / text_w / 2);
    HAdj->set_upper (70.0);
    HAdj->changed ();
  }
}

void MemoryPanel::xChanged (void)
{
  xOffs = (int) HAdj->get_value ();
  queue_draw ();
}

void MemoryPanel::yChanged (void)
{
  yOffs = (int) VAdj->get_value ();
  queue_draw ();
}

void MemoryPanel::showLine (Glib::RefPtr<Gdk::Window> win, unsigned line, int y)
{
Glib::RefPtr<Gdk::GC> gc;
int i;
char buf [80];

   gc = get_style()->get_black_gc();
  if (line < data -> getSize () / 16)
  {
    sprintf (buf,"%04X  ",line * 16 + data -> getStart ());
    for (i = 0; i < 16; i++)
    {
    	sprintf (buf + 5 + i * 3,"%02X ",data -> GetDato (16 * line + i + data -> getStart ()));
    }
    for (i = 0; i < 16; i++)
    {
    	buf [53 + i] = (data->GetDato (16 * line + i + data -> getStart ()) >= ' ' && 
			data -> GetDato (16 * line + i + data -> getStart ()) < 128) ? data -> GetDato (16 * line + i + data -> getStart ()) : '.';
    }
    buf [53 + i] = 0;
    i = (xOffs < 69) ? xOffs : 69;
    Glib::RefPtr<Pango::Layout> pangolayout = create_pango_layout(buf + i);
    pangolayout->set_font_description  (*font); 
    win->draw_layout(gc, border , y, pangolayout);

    // Cursore
    if (curOn && curY == line) {
	  win->draw_rectangle (gc,true,(curPos () - xOffs) * text_w + border,
		y + text_h - 2, text_w, 2);
    }
  }
}

unsigned MemoryPanel::curPos ()
{
  return ((Zone) ? 53 + curX : 5 + curX * 3 + low);
}

bool MemoryPanel::on_expose_event (GdkEventExpose *event)
{
unsigned i; 
Glib::RefPtr<Gdk::Window> win = get_window ();
Glib::RefPtr<Gdk::GC> gc;
    gc = get_style()->get_white_gc();
    win -> draw_rectangle (gc,
			true,
			event->area.x, event->area.y,
			event->area.width, event->area.height);

    for (i = 0; border + text_h * i < (unsigned) get_height (); i ++) {
		if ((unsigned) (event->area.y) <= (border + text_h * (i + 1)) && 
	     ((unsigned) (event->area.y + event->area.height)) >= (border + text_h * i))
    	{
        	showLine (win, i + yOffs, border + text_h * i);
    	}
	}
 
    return false;
}

bool MemoryPanel::on_button_press_event (GdkEventButton *b)
{
  curOn = true;
  tick();
  curY = ((int) b -> y - border)/text_h + yOffs;
  setPosition (((int) b -> x - border)/text_w + xOffs);
  grab_focus();
  return true;
}

void MemoryPanel::setPosition (int x)
{
  if (x < 5)
  {
    curX = 0;
    Zone = 0;
    low = 0;
  }
  else if (x < 53)
  {
    Zone = 0;
    x -= 5;
    low = (x %3) ? 1 : 0;
    curX = x / 3;
  }
  else if (x < 70)
  {
    curX = x-53;
    Zone = true;
  }
  else
  {
    curX = 15;
    Zone = true;
  }
}

bool MemoryPanel::tick()
{
    curOn = !curOn;
    // Cursore
    queue_draw_area ((curPos () - xOffs) * text_w + border, (curY - yOffs) * text_h + border, text_w, text_h);
    return true;
}

bool MemoryPanel::on_focus_in_event (GdkEventFocus *e)
{
    cursorTimer = Glib::signal_timeout().connect(sigc::mem_fun(*this,&MemoryPanel::tick),500);
    curOn = false;	// tick lo inverete, prima di 'rinfrescare' il video
	tick ();		// ridisegna il carattere alla posizione del cursore
	return false;
}

bool MemoryPanel::on_focus_out_event (GdkEventFocus *e)
{
    cursorTimer.disconnect ();
    curOn = true;	// tick lo inverete, prima di 'rinfrescare' il video
	tick ();		// ridisegna il carattere alla posizione del cursore
	return false;
}

void MemoryPanel::key (int k)
{
GdkRectangle redraw_rect, redraw_rect1;
	curOn = true;
	tick();	// Spegne il cursore
	switch (k)
	{
		case GDK_Up :  // Freccia Su
			if (curY) {
				curY--;
			}
			break;
		case GDK_Down:  // Freccia Giu
			if (curY < (data -> getSize () / 16) - 1) {
				curY++;
			}
			break;
		case GDK_Right :  // Freccia Destra
			if (!Zone) {
				low = !low;
			}
			if (!low || Zone) {
				if (++curX > 15) {
					if (curY < (data -> getSize () / 16) - 1) {
						curX = 0;
						curY ++;
					} else {
						curX --;
					}
				}
			}
			break;
		case GDK_Left :  // Freccia Sinistra
			if (!Zone) {
				low = !low;
			}
			if (low || Zone) {
				if (curX-- == 0) {
					if (curY) {
						curX = 15;
						curY --;
					} else {
						curX ++;
					}
				}
			}
			break;
		case GDK_Page_Up :  // Pagina Su
			curY -= (get_height () - 2 * border) / text_h / 2;
			if (curY < 0) {
				curY = 0;
			}
			yOffs -= (get_height () - 2 * border) / text_h / 2;
			if (yOffs < 0) {		  
				yOffs = 0;
			}
			VAdj->set_value (yOffs);
			break;
		case GDK_Page_Down :  // Pagina Giu
			curY += (get_height () - 2 * border) / text_h / 2;
			if (curY >= data -> getSize () / 16) {
				curY = data -> getSize () / 16 - 1;
			}
			yOffs += (get_height () - 2 * border) / text_h / 2;
			if (yOffs >= data -> getSize () / 16 - (get_height () - 2 * border) / text_h)  {
				yOffs = data -> getSize () / 16 - (get_height () - 2 * border) / text_h;
			}
			VAdj->set_value (yOffs);
			break;
		case GDK_Home :  // Home
			curY = 0;
			curX = 0;
			break;
		case GDK_End :  // End
			curY = (data -> getSize () / 16) - 1;
			curX = 15;
			break;
		case GDK_Tab :
			Zone = !Zone;
			break;
		default :
			redraw_rect.x = (5 + curX * 3 - xOffs) * text_w + border;
			redraw_rect.y = (curY - yOffs) * text_h + border;
			redraw_rect.width = 70 * text_w;
			redraw_rect.height = text_h;
			
			redraw_rect1.x = (53 + curX - xOffs) * text_w + border;
			redraw_rect1.y = (curY - yOffs) * text_h + border;
			redraw_rect1.width = text_w;
			redraw_rect1.height = text_h;
			
			if (Zone) {	// Ascii
				data -> SetDato (16 * curY + curX, k);
				data -> saveValue (16 * curY + curX++);
			} else {
				if (k >= '0' && k <= '9') {
					k -= '0';
				} else if (k >= 'a' && k <= 'f') {
					k = k - 'a' + 10;
				} else if (k >= 'A' && k <= 'F') {
					k = k - 'A' + 10;
				} else {
					break;
				}
			
				if (low) {
				unsigned char tmp = data -> GetDato (16 * curY + curX);
					tmp &= 0xf0;
					tmp |= k;
					data -> SetDato (16 * curY + curX, tmp);
					data -> saveValue (16 * curY + curX);
					low = false;
					curX ++;
				} else {
				unsigned char tmp = data -> GetDato (16 * curY + curX);
					tmp &= 0x0f;
					tmp |= k << 4;
					data -> SetDato (16 * curY + curX, tmp);
					data -> saveValue (16 * curY + curX);
					low = true;
				}
			}
			if (curX > 15) {
				if (curY < (data -> getSize () / 16) - 1) {
					curX = 0;
					curY ++;
				} else {
					curX --;
				}
			}
			queue_draw_area (redraw_rect.x, redraw_rect.y, redraw_rect.width, redraw_rect.height);
			queue_draw_area (redraw_rect1.x, redraw_rect1.y, redraw_rect1.width, redraw_rect1.height);
	}
	// Visualizza la posizione del cursore
	if (VAdj) {
		if (curY < yOffs) {
			VAdj->set_value (curY);
		} else if (curY >= yOffs + (get_height () - border)/text_h)  {
			VAdj->set_value (curY - (get_height () - border)/text_h + 1);
		}
	}
	if (HAdj) {
		if (curPos () < xOffs) {
			HAdj->set_value (curPos ());
		} else if (curPos () >= xOffs + (get_width () - border)/text_w) {
			HAdj->set_value (curPos () - (get_width () - border)/text_w + 1);
		}
	}
}

bool MemoryPanel::on_key_press_event (GdkEventKey* p1)
{
char *c = p1->string;
  if (*c)
    while (*c)
    {
      key (*c++);
    }
  else
  {
    key (p1 -> keyval);
  }
  return true;
}

void MemoryPanel::goTo (unsigned where)
{
}
