/***************************************************************************
 *  sfr.cc
 *  Special Function Registers
 *
 *  Sun Nov 14 10:49:20 2004
 *  Copyright  2004 Ing. Stefano Salvi
 *  Email stefano@salvi.mn.it
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 

#include <glib/gmacros.h>
#include <glib.h>
#include <gtkmm.h>
#include <libgnomemm.h>

#include <iostream>	// Per la stampa dell'eventuale errore

#ifndef _DMEMORY_H
#include "dmemory.h"
#endif

#include "pcmon.h"
#include "sfr.h"
#include "pc.h"
#include "sfrnames.h"
#include "ppiwin.h"

#ifndef _I8052_H
#include "i8052.h"
#endif

void Sfr::redraw ()
{
  sp.redraw ();
  a.redraw ();
  b.redraw ();
  psw.redraw ();
  dptr.redraw ();
  r0.redraw ();
  r1.redraw ();
  r2.redraw ();
  r3.redraw ();
  r4.redraw ();
  r5.redraw ();
  r6.redraw ();
  r7.redraw ();
}

void Sfr::toolbarBuild (Pcmon *parent)
{
  //Create actions for menus and toolbars:
  m_refActionGroup = Gtk::ActionGroup::create();

  m_refActionGroup->add( Gtk::Action::create("Esegui", Gtk::Stock::EXECUTE),
  	sigc::mem_fun(*parent,&Pcmon::menuRun) );
  m_refActionGroup->add( Gtk::Action::create("EseguiInizio", Gtk::StockID("pcmon_btn_runstart")),
  	sigc::mem_fun(*parent,&Pcmon::menuRunStart) );
  m_refActionGroup->add( Gtk::Action::create("PassoSingolo", Gtk::StockID("pcmon_btn_singlestep")),
  	sigc::mem_fun(*parent,&Pcmon::menuStep) );

  m_refActionGroup->add( Gtk::Action::create("Ricarica", Gtk::Stock::REFRESH),
  	sigc::mem_fun(*this,&Sfr::menuRefresh) );

  m_refUIManager = Gtk::UIManager::create();
  m_refUIManager->insert_action_group(m_refActionGroup);
  add_accel_group(m_refUIManager->get_accel_group());

  try
  {
    /* Crea l'immagine per tutti i menu (principale e popup) e le toolbar */
    Glib::ustring ui_info =
        "<ui>"
        "  <toolbar  name='ToolBar'>"
        "    <toolitem action='Esegui'/>"
        "    <toolitem action='EseguiInizio'/>"
        "    <toolitem action='PassoSingolo'/>"
        "    <separator/>"
		"    <toolitem action='Ricarica'/>"
        "  </toolbar>"
        "</ui>";

    m_refUIManager->add_ui_from_string(ui_info);
  }
  catch(const Glib::Error& ex)
  {
    std::cerr << "Preparazione dei menu fallita: " <<  ex.what();
  }

  //Get the toolbar widgets, and add them to a container widget:
  Gtk::Toolbar* toolbar = (Gtk::Toolbar *) m_refUIManager->get_widget("/ToolBar") ;
  if(toolbar)
    vbox.pack_start(*toolbar, Gtk::PACK_SHRINK);
  toolbar->set_toolbar_style(Gtk::TOOLBAR_ICONS);
}

Sfr::Sfr (Pcmon *p, i8052 *processor) :
  sp("SP", true, false, processor->GetSFR()->GetSPReg()),
  a("A", true, false,processor->GetSFR()->GetAccReg()),
  b("B", true, false,processor->GetSFR()-> GetBReg()),
  psw("PSW", true, false, processor->GetSFR()-> GetPSWReg()),
  dptr("DPTR", false, false, processor->GetSFR() -> GetDPTRReg()),
  r0("R0", true, false, processor->GetData ()->getGPreg(0)),
  r1("R1", true, false, processor->GetData ()->getGPreg(1)),
  r2("R2", true, false, processor->GetData ()->getGPreg(2)),
  r3("R3", true, false, processor->GetData ()->getGPreg(3)),
  r4("R4", true, false, processor->GetData ()->getGPreg(4)),
  r5("R5", true, false, processor->GetData ()->getGPreg(5)),
  r6("R6", true, false, processor->GetData ()->getGPreg(6)),
  r7("R7", true, false, processor->GetData ()->getGPreg(7)),
  vboxl (false, 5),
  vboxr (false, 5),
  hbox (false, 5)
{
	m = processor->GetData ();
	set_resizable (false);	// Finestra non ridimensionabile
	set_title ("Sfr");

	toolbarBuild (p);

	vbox.pack_start(hbox, false,false,0);
	hbox.pack_start(vboxl, false,false,0);
	hbox.pack_start(vboxr, false,false,0);

	vboxl.pack_start(*sp.getBox (), false,false,0);
	vboxl.pack_start(*a.getBox (), false,false,0);
	vboxl.pack_start(*b.getBox (), false,false,0);
	vboxl.pack_start(*psw.getBox (), false,false,0);
	vboxl.pack_start(*dptr.getBox (), false,false,0);

	vboxr.pack_start(*r0.getBox (), false,false,0);
	vboxr.pack_start(*r1.getBox (), false,false,0);
	vboxr.pack_start(*r2.getBox (), false,false,0);
	vboxr.pack_start(*r3.getBox (), false,false,0);
	vboxr.pack_start(*r4.getBox (), false,false,0);
	vboxr.pack_start(*r5.getBox (), false,false,0);
	vboxr.pack_start(*r6.getBox (), false,false,0);
	vboxr.pack_start(*r7.getBox (), false,false,0);

	add (vbox);
	redraw ();
	show_all ();
}
