/***************************************************************************
 *  sfrmemory.cc
 *
 *  Sun Nov 14 10:49:20 2004
 *  Copyright  2004 Ing. Stefano Salvi
 *  Email stefano@salvi.mn.it
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include "sfrnames.h"
#include "serial.h"
#include "sfrmemory.h"
#include "messageinterface.h"
#ifndef _PPIINTERFACE_H
#include "ppiinterface.h"
#endif

/* Class ppiinterface */
void ppiinterface::Out (int address, int datum, int emulating)
{
}
	
unsigned ppiinterface::In (int address)
{
	return 0;
}

void ppiinterface::P1Out (int datum, int emulating)
{
}

void ppiinterface::redrawIfChanged()
{
}
/* Fine class ppiinterface */

ppiinterface DummyPPI;

extern Updatable dummyU;

/* HRegister */
HRegister::HRegister (int is8Bit, char *c)
{
	is8bit = is8Bit;
	value=0;
	set [0] = c [0];
	set [1] = c [1];
	r = &dummyU;
}

HRegister &HRegister::operator= (int i)
{
	value = i & ((is8bit) ? 0xff : 0xffff);	// Limita a 8/16 bit
	redraw ();
	return *this;
}

void HRegister::store ()
{
char buf [20];
	Serial_Container::s -> schPutc (set [0]);	// Comando
	if (Serial_Container::s -> schWaitc (' ')) {
		return;
	}
	Serial_Container::s -> schPutc (set [1]);  // Registro
	if (Serial_Container::s -> schWaitc (' ')) {
		return;
	}
	sprintf (buf,(is8bit) ? "%02x" : "%04x", value);
	Serial_Container::s -> schPuts (buf);
	Serial_Container::s -> schPrompt ();
}

/* BpHRegister */

HRegister &BpHRegister::operator= (int i)
{
	value = i & ((is8bit) ? 0xff : 0xffff);	// Limita a 8/16 bit
	redraw ();
	return *this;
}

/* BpPool */

bool BpPool::checkBps (unsigned v, int bpindex)
{
int j;

	if (v) {
		for (j = 0; j < NUMBP; j++) {
			if (j != bpindex && v >= (unsigned)BPs[j]->getValue () && v < (unsigned)BPs[j]->getValue () + 3) {
				if (bpindex >= 0) {
					MessageInterface::doYesDialog ("Breakpoint", NULL, 
						"Il Breakpoint %d e' nei tre byte usati dal Breakpint %d", bpindex, j);
				}
				return false;
			}
		}
	}
	return true;
}

bool BpPool::isBp (unsigned v)
{
int j;

	for (j = 0; j < NUMBP; j++) {
		if (v == (unsigned)BPs[j]->getValue ()) {
			return true;
		}
	}
	return false;
}

void BpPool::addBkp (unsigned PC)
{
	if (hasBp ()) {
	int j;
		for (j = 0; j < NUMBP; j++) {
			if (PC >= (unsigned) BPs[j]->getValue () && PC <  (unsigned)BPs[j]->getValue () + 3) {
				MessageInterface::doYesDialog ("Breakpoint" , NULL, "Il breakpoint %d occupa gia' questa locazione", j);
				return;
			}
		}
		for (j = 1; j < NUMBP; j++) {	// lascia libero il BP 0
			if (!BPs[j]->getValue ()) {
      			*BPs[j] = PC;	// Setta breakpoint
				BPs[j]  -> store ();
				return;
			}
		}
		MessageInterface::doYesDialog ("Breakpoint" , NULL, "Non ci sono piu' breakpoint disponibili");
	}
}

void BpPool::delBkp (unsigned PC)
{
int j;
	for (j = 0; j < NUMBP; j++) {
		if ((unsigned)BPs[j]->getValue () == PC) {
			*BPs[j] = 0;	// Resetta breakpoint
			BPs[j]  -> store ();
			return;
		}
	}
}

bool BpPool::hasBp ()
{
int j;

	for (j = 1; j < NUMBP; j++) {	// lascia libero il BP 0
		if (!BPs[j]->getValue ()) {
			return true;
		}
	}
	return false;
}

/*
 * DataMemory
 * Specializza per SFR
 */

SFRMemory::SFRMemory (Serial_Container *c_s) : 
  Memory (128, 128),
  sp(true, "CS"),
  a(true, "CA"),
  b(true, "CB"),
  psw(true, "CW"),
  dptr(false, "CD"),
  pc("CP", -1),
  b0("B0", 0),
  b1("B1", 1),
  b2("B2", 2),
  b3("B3", 3)
{
	// Inizializza i valori, nel caso sia 'offline'
	sp = 0x80;
	a = 0;
	b = 0;
	psw = 0;
	dptr = 0;
	// Collega registri e pool
	pool.setRegisters ( &pc, &b0, &b1, &b2, &b3);
	pc.setPool(&pool);
	b0.setPool(&pool);
	b1.setPool(&pool);
	b2.setPool(&pool);
	b3.setPool(&pool);
	// Prima inizializza i valori, nel caso sia 'offline'
	pc = 0x8050;
	b0 = 0;
	b1 = 0;
	b2 = 0;
	b3 = 0;
	
	s = c_s;
	PPI = &DummyPPI;
}

/* Funzione vuota, perche' viene sostituita dai songoli registri */
void SFRMemory::saveValue (int pos)
{
}

/* Funzione vuota: contentua in "dmemory", perche' associata alla letura dei registri GPR */
void SFRMemory::reload ()
{
}

void SFRMemory::store (int where)
{
int Bp0 = b0.getValue ();
int Bp1 = b1.getValue ();
int Bp2 = b2.getValue ();
int Bp3 = b3.getValue ();
	if (Bp0 && where >= Bp0 && where < Bp0 + 3) {
		MessageInterface::doYesDialog ("Esecuzione", NULL, 
			"Il breakpoint 0 e' nell'area del PC, quindi e' sato omesso");
	}
	if (Bp1 && where >= Bp1 && where < Bp1 + 3) {
		MessageInterface::doYesDialog ("Esecuzione", NULL, 
			"Il breakpoint 1 e' nell'area del PC, quindi e' sato omesso");
	}
	if (Bp2 && where >= Bp2 && where < Bp2 + 3) {
		MessageInterface::doYesDialog ("Esecuzione", NULL, 
			"Il breakpoint 2 e' nell'area del PC, quindi e' sato omesso");
	}
	if (Bp3 && where >= Bp3 && where < Bp3 + 3) {
		MessageInterface::doYesDialog ("Esecuzione", NULL, 
			"Il breakpoint 3 e' nell'area del PC, quindi e' sato omesso");
	}
	pc.store ();
	/* Il Breakpoint 0 (esegui fino...) viene posto in maniera incondizionata, senza
	 * controllare la posizione degli altri breakpoint.
	 * Se il breakpoint 'vai a' collide con un altro breakpoint, viene data la precedenza
	 * al 'vai a', eliminando l'altro.
	 */
	b0.setValue ((Bp0 && where >= Bp0 && where < Bp0 + 3) ? 0 : Bp0);
	if (Bp1 && ((where >= Bp1 && where < Bp1 + 3) || (Bp0 && Bp0 + 3 > Bp1 && Bp0 < Bp1 + 3))) {
		b1.setValue (0);
		b1 = Bp1;
	} else {
		b1.store ();
	}
	if (Bp2 && ((where >= Bp2 && where < Bp2 + 3) || (Bp0 && Bp0 + 3 > Bp2 && Bp0 < Bp2 + 3))) {
		b2.setValue (0);
		b2 = Bp2;
	} else {
		b2.store ();
	}
	if (Bp3 && ((where >= Bp3 && where < Bp3 + 3) || (Bp0 && Bp0 + 3 > Bp3 && Bp0 < Bp3 + 3))) {
		b3.setValue (0);
		b3 = Bp3;
	} else {
		b3.store ();
	}
	sp.store ();
	a.store ();
	b.store ();
	psw.store ();
	dptr.store ();
}

unsigned char SFRMemory::GetSFR (int reg)
{
  switch (reg)
  {
    case aSP:
      return sp.getValue ();
    case aDPL:
      return dptr.getValue () & 0xff;
    case aDPH:
      return (dptr.getValue () >> 8) & 0xff;
    case aPSW:
      return psw.getValue ();
    case aACC:
      return a.getValue ();
    case aB:
      return b.getValue ();
  }
  return data [reg & 0x7f];
}

void SFRMemory::SetSFR (int reg,unsigned v)
{
  switch (reg & 0xff)
  {
    case aSP:
      sp = v;
      break;
    case aDPL:
      dptr = (dptr.getValue () & 0xff00) | (v & 0xff);
      break;
    case aDPH:
      dptr = (dptr.getValue () & 0xff) | ((v << 8) & 0xff00);
      break;
    case aPSW:
      psw = v;
      break;
    case aACC:
      a = v;
      break;
    case aB:
      b = v;
      break;
    case aSBUF:
      SendSerial = 10;
      s -> tPutc (v);
      break;
    case aP1:
      PPI -> P1Out (v, status == emulating);	// Per i due LED su P1.2 e P1.5
    default :
      data [reg & 0x7f] = v;
      break;
  }
}

void SFRMemory::setStatus (stat_val s)
{
	Memory::setStatus (s);
	b1.setStatus(s);
	b2.setStatus(s);
	b3.setStatus(s);
	pc.setStatus(s);
	sp.setStatus(s);
	a.setStatus(s);
	b.setStatus(s);
	psw.setStatus(s);
	dptr.setStatus(s);
}

void SFRMemory::redraw ()
{
	b1.redraw();
	b2.redraw();
	b3.redraw();
	pc.redraw();
	sp.redraw();
	a.redraw();
	b.redraw();
	psw.redraw();
	dptr.redraw();
}

void SFRMemory::redrawIfChanged ()
{
	b1.redrawIfChanged();
	b2.redrawIfChanged();
	b3.redrawIfChanged();
	pc.redrawIfChanged();
	sp.redrawIfChanged();
	a.redrawIfChanged();
	b.redrawIfChanged();
	psw.redrawIfChanged();
	dptr.redrawIfChanged();
	PPI->redrawIfChanged();
}

void SFRMemory::receiver (char c)
{
	data [aSBUF & 0x7f] = c;
	data [aSCON & 0x7f] |= 1;	// Accende RI
}
