/***************************************************************************
 *  termdrawing.cc
 *  Stringa per test del modem : at&v
 *
 *  Sun Nov 14 10:49:20 2004
 *  Copyright  2004 Ing. Stefano Salvi
 *  Email stefano@salvi.mn.it
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include <glib/gmacros.h>
#include <glib.h>
#include <gtkmm.h>
#include <libgnomemm.h>

#include "termdrawing.h"

bool terminal_drawing_area::on_expose_event (GdkEventExpose *event)
{
int i,j; 
char linebuf [nChars + 1];
Glib::RefPtr<Gdk::Window> win;
Glib::RefPtr<Gdk::GC> gc;

    win = get_window();
    gc = get_style()->get_white_gc();
    // Same like above, + Gtk::Widget has set_state function but no get_state 
    // function. 
    win->draw_rectangle (gc,
			true,
			event->area.x, event->area.y,
			event->area.width, event->area.height);

    gc = get_style()->get_black_gc();
	Glib::RefPtr<Pango::Layout> pangolayout = create_pango_layout("");
	pangolayout->set_font_description  	(*font); 
    for (i = 0; i < nLines; i ++) {
	  if (event->area.y <= (border + text_h * i) && 
	      (event->area.y + event->area.height) >= (border + text_h * (i + 1)))
      {
		  for (j = 0; j < nChars; j++) {
			  linebuf [j] = page [i][j];
		  }
		  linebuf [j] = 0;
		  pangolayout -> set_text (linebuf);
		  win->draw_layout(gc, border , border + /* text_a + */ text_h * i, pangolayout);
      }
    }

    // Cursore
    if (curOn) {
  	  win->draw_rectangle (gc,true,curX * text_w + border,
		(curY + 1) * text_h + border - 2, text_w, 2);
    }
    return true;
}

bool terminal_drawing_area::tick() {
    curOn = !curOn;
    queue_draw_area (curX * text_w + border, curY * text_h + border, text_w, text_h);

    return true;
}

bool terminal_drawing_area::gotFocus (GdkEventFocus *e)
{
    cursorTimer = Glib::signal_timeout().connect(sigc::mem_fun(*this,&terminal_drawing_area::tick),500);
    curOn = false;	// tick lo inverete, prima di 'rinfrescare' il video
	tick ();		// ridisegna il carattere alla posizione del cursore
	return false;
}

bool terminal_drawing_area::lostFocus (GdkEventFocus *e)
{
    cursorTimer.disconnect ();
    curOn = false;	// tick lo inverete, prima di 'rinfrescare' il video
	tick ();		// ridisegna il carattere alla posizione del cursore
	return false;
}

void terminal_drawing_area::clearLine (char *line)
{
int j;
    for (j = 0; j < nChars; j ++)
	line [j] = ' ';
}

void terminal_drawing_area::scroll ()
{
GdkRectangle redraw_rect;
char *l;
int i;

    l = page [0];
    clearLine (l);
    for (i = 0; i < nLines - 1; i ++)
	page [i] = page [i + 1];
    page [i] = l;
    if (curY)
	curY --;

    redraw_rect.x = 0;
    redraw_rect.y = 0;
    redraw_rect.width = nChars * text_w + border * 2;
    redraw_rect.height = nLines * text_h + border * 2;
    queue_draw ();
}

void terminal_drawing_area::tputc (char c)
{
	// Ridisegno il posto dove scrivero'
    queue_draw_area (curX * text_w + border, curY * text_h + border, text_w, text_h);
    if (c == 13)	// CR
    {
	  curX = 0;
    } else if (c == 10)	// LF
    {
	  curY ++;
	  if (curY >= nLines) {
	    scroll ();
	  }
    } else if (c >= ' ')	 {	// Essendo 'signed', non controllo il massimo
	  page [curY][curX] = c;
	  curX ++;
	  if (curX >= nChars)
	  {
	    curX = 0;
	    tputc (10);	// Stampa un 'LF'
	  }
    }

    curOn = 0;
}

terminal_drawing_area::terminal_drawing_area ()
{
int i;
    curOn = -1;
    page = new char* [nLines];
    for (i = 0; i < nLines; i++)
    {
		page [i] = new char [nChars];
		clearLine (page [i]);
    }
    curX = curY = 0;
 	
	font = new Pango::FontDescription ("Courier,monospace, 10");
	Glib::RefPtr<Pango::Layout> pangolayout = create_pango_layout("M");
	pangolayout->set_font_description (*font);
	pangolayout->get_pixel_size (text_w, text_h);

    set_size_request ( nChars * text_w + border * 2, nLines * text_h + border * 2);

    set_events (Gdk::EXPOSURE_MASK);
}
